//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalAtomsMisc.cc --- implementation of miscallaneous CalAtomsCommon classes
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1997-07-25: 1.0.0: - skeleton
//		UL, 1997-??-??: 1.1.0: - extended
//		UL, 1998-08-28: 1.2.0: - EnergyCorrector added
//		UL, 1999-03-05: 1.3.0: - changes enforced by lowmem Dal
//		UL, 1999-05-11: 1.4.0: - channel range of Ebounds/Ebins is now
//								 0-4095 for EMOS+EPN
//							   - changed fatal -> error and changed all
//								 error tags
//		UL, 2000-11-09: 1.5.0: - changes to cope with vector columns in
//								 CCF::MODEPARAM; this fixes SPR 1827:
//								 Does not recognize PrimeFullWindow for MOS 1
//		UL, 2001-02-19: 1.6.0: - moved BadPixel related stuff into separate
//								 file
//		UL, 2002-06-14: 1.7.0: - moved Background related stuff into separate
//								 file
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/CalAtomsMisc.cc,v 1.8 2003/10/03 14:15:40 sasbuild Exp $
//

#include "CalAtomsMisc.h"
#include "CalOalUtility.h"
#include "CalServer.h"

#include <algorithm>
#include <cmath>

using namespace CalOalUtility;
using namespace std;

//
//	----------------------( mode parameters )-----------------------------
//
CalAtomCollection<ModeParameters>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nModes		= instrument->nModes();
	_nEntries	= _nModes;
	ModeParameters *null = 0;
	_collection = new vector<ModeParameters *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::ModeParam);
	DataSet *dataSet = e->open();
	Table *table = dataSet->table(
					e->queryBlockName(Ccf::toString(Ccf::ModeParam)));

	ColumnData *cdmodeId = table->column("MODE_ID")->data(),
			   *cdintTime = table->column("INTEGRATION_T")->data(),
			   *cdshiftTime = table->column("SHIFT_T")->data(),
			   *cdcycleTime = table->column("CYCLE_T")->data(),
			   *cdtimeRes = table->column("TIME_RES")->data(),
			   *cddutyCycle = table->column("DUTY_CYCLE")->data(),
			   *cdsmearedFrac = table->column("SMEARED_FRAC")->data(),
			   *cdpntSrcPileUp = table->column("PTSRCE_PILEUP")->data(),
			   *cdlimitingFlux = table->column("LIMITING_FLUX")->data(),
			   *cdwx0 = table->column("WINDOWX0")->data(),
			   *cdwy0 = table->column("WINDOWY0")->data(),
			   *cdwdx = table->column("WINDOWDX")->data(),
			   *cdwdy = table->column("WINDOWDY")->data();
	int8	*modeId = cdmodeId->int8Data();
	real32	*intTime = cdintTime->real32Data(),
			*shiftTime = cdshiftTime->real32Data(),
			*cycleTime = cdcycleTime->real32Data(),
			*timeRes = cdtimeRes->real32Data(),
			*dutyCycle = cddutyCycle->real32Data(),
			*smearedFrac = cdsmearedFrac->real32Data(),
			*pntSrcPileUp = cdpntSrcPileUp->real32Data(),
			*limitingFlux = cdlimitingFlux->real32Data();
	int16	*wx0 = cdwx0->int16Data(),
			*wy0 = cdwy0->int16Data(),
			*wdx = cdwdx->int16Data(),
			*wdy = cdwdy->int16Data();

	short underX = table->attribute("UNDER_X")->asInt(),
		  overX = table->attribute("OVER_X")->asInt(),
		  underY = table->attribute("UNDER_Y")->asInt(),
		  overY = table->attribute("OVER_Y")->asInt();

	TimeCorrector *tc = 0; tc = calServer.getAtom(tc);
	vector<double> ft = tc->frameTimes();
	for (unsigned n=0; n<table->rows(); ++n) {
		if (modeId[n] > INT(_nModes))
			errstr<<"Illegal mode number in row #"
				  <<n<<" of CCF file "
				  <<Xmm::toString(instrument->id())<<"/"
				  <<Ccf::toString(Ccf::ModeParam)
			  <<Sas::error("IllegalValues");

		unsigned pos = index(modeId[n]);

		if (!(*_collection)[pos])
			(*_collection)[pos] = new ModeParameters;

		(*_collection)[n] = new ModeParameters;

		#define VALUE(a)	a[cd##a->position(n, 0)]

		(*_collection)[n]->_integrationTime = VALUE(intTime);
		(*_collection)[n]->_shiftTime = VALUE(shiftTime);
		(*_collection)[n]->_frameTime = n < ft.size() ? ft[n] : 0.;
		(*_collection)[n]->_cycleTime = VALUE(cycleTime);
		(*_collection)[n]->_timeResolution = VALUE(timeRes);
		(*_collection)[n]->_dutyCycle = VALUE(dutyCycle);
		(*_collection)[n]->_smearedFraction = VALUE(smearedFrac);
		(*_collection)[n]->_pointSourcePileUp = VALUE(pntSrcPileUp);
		(*_collection)[n]->_limitingFlux = VALUE(limitingFlux);
		(*_collection)[n]->_windowX0 = VALUE(wx0);
		(*_collection)[n]->_windowY0 = VALUE(wy0);
		(*_collection)[n]->_windowDX = VALUE(wdx);
		(*_collection)[n]->_windowDY = VALUE(wdy);
		(*_collection)[n]->_underX = underX;
		(*_collection)[n]->_overX = overX;
		(*_collection)[n]->_underY = underY;
		(*_collection)[n]->_overY = overY;

		#undef VALUE
	}
	delete cdmodeId;
	delete cdintTime;
	delete cdshiftTime;
	delete cdcycleTime;
	delete cdtimeRes;
	delete cddutyCycle;
	delete cdsmearedFrac;
	delete cdpntSrcPileUp;
	delete cdlimitingFlux;
	delete cdwx0;
	delete cdwy0;
	delete cdwdx;
	delete cdwdy;
	dataSetServer->close(dataSet);
} // CalAtomCollection<ModeParameters>::CalAtomCollection


ModeParameters *CalAtomCollection<ModeParameters>::select(
					Ccd,
					Node,
					unsigned modeOffset) const
{
	unsigned n = index(modeOffset);

	return (*_collection)[n];
} // CalAtomCollection<ModeParameters>::select
