//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalServer.h	--- the top-level CAL include file
//
//	DESCRIPTION:
//		CalServer is a central class with functionality to instantiate
//		and access CAL atoms; there is a singleton CalServer that gets
//		instantiated when first time used; the CalServer has a state
//		that can be set from various inputs (see setState() methods below)
//		and after that accessed (getState()). The CalState class provides
//		methods to get and set individual state variables. CalStates can
//		be pushed (pushState()) and poped (popState()) onto/from a LIFO
//		stack. When instaniated the CalServer also tries to access a valid
//		CCF (via environment variable SAS_CCF); this can be changed
//		with setCcf(); access to the CCF can be obtained with ccf();
//		The instantiation of new CAL atoms is logged in a CcfAccessLog
//		access to which is provided through CalServer::ccfAccessLog().
//		There is a central random number generator that is used througout
//		the CAL - it can be controlled via CalServer::setRandom().
//
//	CHANGE HISTORY:
//		UL, 1998-04-22: 1.0.0: - first version
//		UL, 1998-11-18: 2.0.0: - overhaul
//		UL, 1999-03-29: 3.0.0: - another overhaul; methods documented
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/CalServer.h,v 1.30 2004/08/04 11:59:08 rsaxton Exp $
//
#ifndef CALSERVER_H
#define CALSERVER_H

#include <string>

#include "Xmm.h"			//	payload characteristics
#include "Ccf.h"			//	CCF stuff
#include "CalAtomsBase.h"	//	Cal Atoms
#include "CalAtomServer.h"	//	atom server
#include "CalState.h"		//	Cal state
#include "CalDataVector.h"	//	data vectors
#include "Dal.h"			//	for inititalizing state from data set

#include <iosfwd>

class CalServer {
public:
	CalServer(Xmm::Instrument instrument = Xmm::EMOS1);
		//
		//	PURPOSE:
		//		Construct a call server object
		//
		//	ARGUMENTS:
		//		In:
		//			instrument: a CalServer object is instrument specific;
		//						the current instrument can be changed again
		//						at any time via setInstrument() below
		//


	void setState(const std::string &dataSet, bool silent = false,
				  bool ignoreFilter = false);
	void setState(const DataSet &dataSet, bool silent = false,
				  bool ignoreFilter = false);
	void setState(const Block &block, bool silent = false,
				  bool ignoreFilter = false);
	void setState(const std::vector<const Attributable *> &av,
				  bool silent = false,
				  bool ignoreFilter = false,
				  const Table *evTable = 0);
		//
		//	PURPOSE:
		//		Define the state and other components of this CalServer
		//		object from the contents of a Block in a Dal data set.
		//		In the former form all blocks of the specified data set
		//		are inspected for the informations which are needed to
		//		initialize the CalServer components (instrument, state)
		//		If the specified data set does not contain all the
		//		required information to re-initialize the object, a
		//		fatal error condition will be raised.
		//


	void setInstrument(Xmm::Instrument instrument);
		//
		//	PURPOSE:
		//		Change the current instrument
		//
		//	ARGUMENTS:
		//		In:
		//			instrument: the instrument to change to; several components
		//						will be internally updated as a consequence
		//						of which previously returned pointers by
		//						CalServer methods will become stale (e.g.
		//						CalServer::state()
		//


	const XmmInstrument *instrument() const { return _instrument; }
		//
		//	PURPOSE:
		//		Return pointer to XmmInstrument object corresponding
		//		to current instrument setting. The XmmInstrument::id()
		//		method shall be used to infer the associated Id (type of
		//		argument for constructor or CalServer::setInstrument()
		//


	CalState &getState() const;
		//
		//	PURPOSE:
		//		Obtain a reference to the current CalState object
		//


	void setState(const CalState &calState);
		//
		//	PURPOSE:
		//		Set the state of the CalServer
		//
		//	ARGUMENTS:
		//		In:
		//			calState: a reference to a CalState object to which
		//					  the CalServer's state is to be set
		//


	void pushState();
		//
		//	PURPOSE:
		//		Push a copy of the current CalState onto a LIFO stack
		//


	void popState();
		//
		//	PURPOSE:
		//		Counterpart to CalServer::pushState() - get the topmost
		//		CalState from the LIFO stack and make it the new state
		//		thereby loosing the current state settings; attempting
		//		to pop more states than have been previously pushed leads
		//		to warning
		//


	CalState *state() const;
		//
		//	PURPOSE:
		//		Obtain a pointer to the current state for accessing
		//		state variables (get/set values)
		//


	Ccf *setCcf(const std::string &cifName = "") const;
		//
		//	PURPOSE:
		//		Construct a new Ccf instance and return pointer to it;
		//		Beware: After CalServer::setCcf() pointers previously
		//		returned by CalServer::ccf() will be stale
		//
		//
		//	ARGUMENTS:
		//		In:
		//			cifName: The name of a directory containing a CCF
		//					 with a CIF named 'ccf.cif' or the name
		//					 of the CIF (incl. path) directly; a fatal
		//					 error condition will be raised in case
		//					 the specified directory/CIF does not exists
		//					 or is unreadable, or the CIF has not the
		//					 right structure.
		//


	Ccf *ccf() const;
		//
		//	PURPOSE:
		//		Obtain pointer to current CCF in use - beware: pointer
		//		will be stale after call to CalServer::setCcf()
		//


	CcfAccessLog &ccfAccessLog() { return CcfAccessLog::instance(); }
		//
		//	PURPOSE:
		//		Give access to singleton CcfAccessLog - this contains
		//		all the CCF constituents that were accessed during the
		//		livetime of this server or since the last reset respectivley
		//


	void setRandom(double min, double max);
	void setRandom();
		//
		//	PURPOSE:
		//		Define interval for random number generator;
		//		Second form restores default values
		//


	CalAtom *getAtom(Ccf::CalDataTypes type);
	template<class AtomType>
	AtomType *getAtom(AtomType *atom, bool generateError = true) const;
	template <class AtomType>
		//
		//	PURPOSE:
		//		Get a particular calibration atom according to the current
		//		state settings - Cal atoms will be initialized from
		//		corresponding CCF constituents
		//
		//	ARGUMENTS:
		//		In:
		//			atom: a pointer to a CalAtom which shall be retrieved;
		//				  example usage:
		//					BadPixelMap *badPixelMap =
		//									calServer.getAtom(badPixelMap);
		//


	void releaseAtom(AtomType *atom) const;
		//
		//	PURPOSE:
		//		Free up resources associated with a particular Cal atom
		//		which has been constructed via CalServer::getAtom()
		//		above
		//
		//	ARGUMENTS:
		//		In:
		//			atom: pointer to Cal atom returned by a previous call
		//				  to CalServer::getAtom; example usage:
		//					calServer.releaseAtom(badPixelMap);
		//


	CalAtomCollectionServer *atomServer() const;
		//
		//	PURPOSE:
		//		Obtain a pointer to the current Cal atom server; beware: after
		//		changes of the current instrument via CalServer::setInstrument
		//		all pointers previously returned by CalServer::atomServer()
		//		will be stale.
		//


	void print(std::ostream &os) const;
		//
		//	PURPOSE:
		//		Print a text representation of the current CalServer onto
		//		a specified ostream.
		//
		//	ARGUMENTS:
		//		In:
		//			os: ostream to print to
		//


	CalInt16Vector int16Vector(const int16 *ptr = 0, unsigned n = 0);
	CalReal32Vector real32Vector(const real32 *ptr = 0, unsigned n = 0);
		//
		//	PURPOSE:
		//		Contruct a container object for a pointer to an int16/real32
		//		data vector; int16 vectors can be converted to real32
		//		vectors via the toReal() method taking into account
		//		randomization; real32 - can be converted into
		//		int16 vectors via toInt()
		//
		//	ARGUMENTS:
		//		In:
		//			ptr	: pointer to memory of suitable size
		//			n	: size of the vector
		//


	~CalServer();
		//
		//	PURPOSE:
		//		Destroy the CalServer object constructed with (any of) above
		//		constructor(s).
		//

private:
	mutable XmmInstrument	*_instrument;		//	the instrument;	
	mutable CalState		*_state;			//	the state
	mutable Ccf				*_ccf;				//	the CCF in use
	mutable	CalAtomCollectionServer
							*_atomServer;		//	the atom server
	double	_oldRandMin, _oldRandMax;			//	old limits of RNG

	std::vector<CalState *>	_stateStack;		//	a state stack

	void defState();
	void checkCcf() const;
	void init();
	std::vector<CoreModeParameters>	coreModeParameters();
}; // CalServer

//
//	initialize CalServer from contents of Dal data set
//
CalServer &operator<<(CalServer &server, const std::string &dataSet);
CalServer &operator<<(CalServer &server, const DataSet &dataSet);
CalServer &operator<<(CalServer &server, const Block &block);
CalServer &operator<<(CalServer &server,
					  const std::vector<const Attributable *> av);

//
//	print CalServer object onto an ostream
//
std::ostream &operator<<(std::ostream &os, CalServer &server);

//
//	there is a single global instance of a CAL server called
//
//		calServer
//
#define calServer   calServerFunc()
CalServer &calServerFunc();

#endif
