//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CrossPsf.cc --- implementation of CrossPsf classes
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-09-16: 1.0.0: - first incarnation
//		UL, 1999-11-08: 1.1.0: - use 'beta' as independent variable on
//								 curves _lorenzianWidth/... instead of -Z
//		UL, 1999-11-15: 1.2.0: - bug fix in RgsCrossPsf::probability
//		UL, 1999-11-25: 1.3.0: - changes I/F + implementation of
//								 CrossPsf::probability() - now receives
//								 vector with cross-angle values
//		UL, 2000-03-03: 1.4.0: - RgsCrossPsf::probability(): center channel
//								 is _negative_ of input source off-axis angle
//		UL, 2000-04-04: 1.4.1: - converted warning into Verbose message
//								 in RgsCrossPsf::probability()
//		UL, 2000-11-23: 1.5.0: - new implementation of
//									RgsCrossPsf::probability()
//								 in line with description of issue 2.1 of [2]
//								 changes are backwards compatible, i.e.,
//								 old CCF::CrossPsf can still be read
//		UL, 2001-12-07: 1.6.0: - changes to normalizations of CrossPsf
//								 components; resulting probabilty spectrum
//								 is no longer renormalized to X-disp window
//								 (see email John Peterson, 2001-12-06)
//		UL, 2002-02-15: 1.7.0: - fixed sign bug in construction of distribution
//
//	AUTHOR:
//		U. Lammers, Thu Sep 16 17:42:38 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.
//		[2]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/CrossPsf.cc,v 1.25 2003/06/16 08:40:14 gvacanti Exp $
//

#include <cmath>

#include "Tracer.h"
#include "CalServer.h"
#include "MiscDataServer.h"
#include "CrossPsf.h"

#include "CalOalUtility.h"
#include "slatec++.h"		//	erf

using namespace CalOalUtility;
using namespace std;

//
//	---( CanonicalCrossPsf )--------------------------------------------------
//
CanonicalCrossPsf::CanonicalCrossPsf(CalAtomCollection<CanonicalCrossPsf>
									*parent)
	: CalChild<CalAtomCollection<CanonicalCrossPsf> >(parent)
{
} // CanonicalCrossPsfCanonicalCrossPsf()


void CanonicalCrossPsf::dump(ostream &)
{
	errstr<<"Cannot dump CanonicalCrossPsf object"<<Sas::error("NoData");
} // CanonicalCrossPsf::dump()


Spectrum CanonicalCrossPsf::probability(double, double, double,
										const CalReal64Vector &, double) const
{
	errstr<<"CanonicalCrossPsf cannot give a cross PSF probability "
			"distribution"<<Sas::error("NoData");

	return Spectrum();
} // CanonicalCrossPsf::probability()


CanonicalCrossPsf::~CanonicalCrossPsf()
{
} // CanonicalCrossPsf::~CanonicalCrossPsf()



CalAtomCollection<CanonicalCrossPsf>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	if (instrument->id() != Xmm::RGS1 &&
		instrument->id() != Xmm::RGS2)
		errstr<<"No CrossPsf information available for instrument "
			  <<Xmm::toString(instrument->id())
			  <<Sas::error("NoSuchInformation");

	_filter		= filter;
	_accuracy	= accuracy;
	_nEntries	= 1;	//	for the time being
	CanonicalCrossPsf *null = 0;
	_collection	= new vector<CanonicalCrossPsf *>(_nEntries, null);

	CanonicalCrossPsf *&cr = (*_collection)[0];

	//
	//	get parameters from CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::CrossPSF);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet, e->queryBlockName("CROSSPSF"));
	const Table *table = ccftab.block();
	cr = new RgsCrossPsf(table, instrument->maxY(), this);

	dataSetServer->close(dataSet);
} // CalAtomCollection<CanonicalCrossPsf>::CalAtomCollection()


CanonicalCrossPsf *CalAtomCollection<CanonicalCrossPsf>::select(Ccd, Node,
																unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<CanonicalCrossPsf>::select()


//
//	---( RgsCrossPsf )--------------------------------------------------------
//
RgsCrossPsf::RgsCrossPsf(const Table *crossPsfTable,
						 unsigned channels,
						 CalAtomCollection<CanonicalCrossPsf> *parent)
		:  CalChild<CalAtomCollection<CanonicalCrossPsf> >(parent),
		   CanonicalCrossPsf(parent),
		  _channels1x1(channels)
{
	//
	//	get some needed misceallaneous data
	//
	MiscDataServer *mds = 0; mds = calServer.getAtom(mds);
	_alphaIncidence = mds->get("INCIDENCE_ANGLE");
	_gammaDesign = _alphaIncidence + mds->get("BLAZE_ANGLE");
	_rowlandRadius = mds->get("ROWLAND");
	_pixelSize1x1 = mds->get("MM_PER_PIXEL_Y");
	_focalLength = mds->get("FOCAL_LENGTH");

	//
	//	read crossPsf table
	//
	errstr<<"Reading cross dispersion parameter from "
		  <<crossPsfTable->qualifiedName()
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	vector<double> beta;
	CalOalUtility::fillVector(crossPsfTable, "BETA", beta);


	vector<vector<double> >	gNorm, gCenter, gWidth;
	CalOalUtility::fillMatrix(crossPsfTable, "GAUSS_SIGMA", gWidth);
	CalOalUtility::fillMatrix(crossPsfTable, "GAUSS_NORM", gNorm);
	if (crossPsfTable->hasColumn("GAUSS_CENTER"))
		//
		//	pre-issue-2 CCFs have no GAUSS_CENTER/LORENTZ_CENTER columns
		//
		CalOalUtility::fillMatrix(crossPsfTable, "GAUSS_CENTER", gCenter);

	unsigned nG = gWidth.size();
	for (unsigned n=0; n<nG; ++n)
		if (gWidth[n][0] != 0. &&
			gNorm[n][0] != 0.) {
			_gWidth.push_back(InterpolatedCurve(beta, gWidth[n]));
			_gNorm.push_back(InterpolatedCurve(beta, gNorm[n]));
			_gCenter.push_back(gCenter.size() > n
									? InterpolatedCurve(beta, gCenter[n])
									: InterpolatedCurve(0.));
		}

	vector<vector<double> >	lNorm, lCenter, lWidth;
	CalOalUtility::fillMatrix(crossPsfTable, "LORENTZ_WIDTH", lWidth);
	CalOalUtility::fillMatrix(crossPsfTable, "LORENTZ_NORM", lNorm);
	if (crossPsfTable->hasColumn("LORENTZ_CENTER"))
		CalOalUtility::fillMatrix(crossPsfTable, "LORENTZ_CENTER", lCenter);
	unsigned nL = lWidth.size();
	for (unsigned n=0; n<nL; ++n)
		if (lWidth[n][0] != 0. &&
			lNorm[n][0] != 0.) {
			_lWidth.push_back(InterpolatedCurve(beta, lWidth[n]));
			_lNorm.push_back(InterpolatedCurve(beta, lNorm[n]));
			_lCenter.push_back(lCenter.size() > n
									? InterpolatedCurve(beta, lCenter[n])
									: InterpolatedCurve(0.));
		}
} // RgsCrossPsf::RgsCrossPsf()


void RgsCrossPsf::dump(ostream &os)
{
	os<<"RgsCrossPsf{\n";
	os<<"Gaussians:\n";
	unsigned nG = _gWidth.size();
	for (unsigned n=0; n<nG; ++n) {
		os<<n<<": width = "<<_gWidth[0](0.)<<endl;
		os<<n<<": norm  = "<<_gNorm[0](0.)<<endl;
		os<<n<<": center= "<<_gCenter[0](0.)<<endl;
	}
	os<<"Lorentzians:\n";
	unsigned nL = _lWidth.size();
	for (unsigned n=0; n<nL; ++n) {
		os<<n<<": width = "<<_lWidth[0](0.)<<endl;
		os<<n<<": norm  = "<<_lNorm[0](0.)<<endl;
		os<<n<<": center= "<<_lCenter[0](0.)<<endl;
	}
	os<<"}\n";
} // RgsCrossPsf::dump()


Spectrum RgsCrossPsf::probability(double, double beta, double sourceXdisp,
								  const CalReal64Vector &xAngles,
								  double binWidth) const
{
	errstr<<"<time> argument of RGS Cross Dispersion PSF construction "
			"currently ignored"<<Sas::message(Msg::Lib, Msg::Verbose);

	double	centerChan = -sourceXdisp;
	vector<double>	norm[2], center[2], width[2];
	//	index 0: Gaussian
	//		  1: Lorentzian

	unsigned nG = _gNorm.size();
	for (unsigned n=0; n<nG; ++n) {
		norm[0].push_back(_gNorm[n](beta));
		center[0].push_back(_gCenter[n](beta));
		width[0].push_back(_gWidth[n](beta));
	}
	unsigned nL = _lNorm.size();
	for (unsigned n=0; n<nL; ++n) {
		norm[1].push_back(_lNorm[n](beta));
		center[1].push_back(_lCenter[n](beta));
		width[1].push_back(_lWidth[n](beta));
	}

	unsigned N[2] = { norm[0].size(), norm[1].size() };
	//	re-normalize Gaussians - with this the relative weight of each
	//	component becomes proportional to the area under the respective
	//	curve
	for (unsigned n=0; n<N[0]; ++n)
		norm[0][n] *= M_SQRT_TWO_PI * width[0][n];

	//	re-normalize Lorentzians - multiply by width to make relative
	//	weight comparable to Gaussian parts
//	for (unsigned n=0; n<N[1]; ++n)
//		norm[1][n] *= width[1][n];

	//	total normalizations
	double tot[2], sum = 0.;
	for (unsigned i=0; i<2; ++i) {
		tot[i] = 0.;
		for (unsigned n=0; n<N[i]; tot[i] += norm[i][n++]);
		sum += tot[i];
	}

	//	calculate first moment and renormalize individual amplitudes
	double firstMoment[2];	//	in beta space
	for (unsigned i=0; i<2; ++i) {
		firstMoment[i] = 0.;
		for (unsigned n=0; n<N[i]; ++n) {
			firstMoment[i] += norm[i][n]*center[i][n]/tot[i];
			norm[i][n] /= sum;
		}
	}

	SAS_DEBUG(
	cout<<"Beta="<<beta<<" rad\n";
	cout<<"First moment of Gauss/Lorentz-components: "
		<<firstMoment[0]<<'/'<<firstMoment[1]<<endl;
	cout<<"Relative weights:\n";
	double snorm = 0.;
	cout<<"Gauss  : "; for (unsigned n=0; n<N[0]; ++n) { snorm += norm[0][n];
							cout<<norm[0][n]<<' '; }
	cout<<endl;
	cout<<"Lorentz: "; for (unsigned n=0; n<N[1]; ++n) { snorm += norm[1][n];
							cout<<norm[1][n]<<' '; }
	cout<<"\nSum: "<<snorm<<endl;
	)

	//
	//	construct the distribution
	//
	unsigned nXDisp = xAngles.size();
	Spectrum spec(nXDisp);
	for (unsigned j=0; j<nXDisp; ++j) {
		spec[j] = 0.;
		//	add up Gaussians
		for (unsigned n=0; n<N[0]; ++n) {
			double oneOversq2sigma = 1./(M_SQRT_2*width[0][n]),
				   argPlus = xAngles[j] + .5 * binWidth - centerChan -
								center[0][n],
				   argMinus = xAngles[j] -.5 * binWidth - centerChan -
								center[0][n];
			spec[j] += .5*norm[0][n] *
						fabs(Slatec::erf(oneOversq2sigma * argPlus) -
							 Slatec::erf(oneOversq2sigma * argMinus));
		}

		//	add up Lorentzians
		for (unsigned n=0; n<N[1]; ++n) {
			double normOverPi = norm[1][n]/M_PI,
				   argPlus = xAngles[j] + .5 * binWidth - centerChan -
								center[1][n],
				   argMinus = xAngles[j] - .5 * binWidth - centerChan -
								center[1][n];
			spec[j] += normOverPi *	fabs(atan2(argPlus, width[1][n]) -
										 atan2(argMinus, width[1][n]));
		}
	}

	//	set first moment
	double fm = (tot[0]*firstMoment[0]+tot[1]*firstMoment[1])/sum;
	spec.firstMoment(fm);

	SAS_DEBUG(
	cout<<"First moment of total distribution: "<<fm<<endl;	
	cout<<"Normalization of CrossPsf in X-disp window: "<<spec.norm()
		<<endl;
	)

	return spec;
} // RgsCrossPsf::probability()


RgsCrossPsf::~RgsCrossPsf()
{
} // RgsCrossPsf::~RgsCrossPsf()
