//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CtiCorrector.cc --- implementation of CtiCorrector class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-06-17: 1.0.0: - first version (extracted from
//								 CalAtomsCommon.cc)
//		UL, 1999-07-01: 1.0.1: - added newline to error message about
//								 failed CTI correction
//		UL, 1999-09-10: 1.1.0: - added RGS CTI correction implementation
//		UL, 1999-09-14: 1.1.1: - incorporated E_PER_ADU factor into
//								 EPN CTI correction
//		UL, 1999-11-24: 1.2.0: - brought in line with calpnalgo-2.0:
//								 pass CCD mode + temperature to CTE init
//								 routine
//		UL, 2000-02-01: 1.7.0: - made EPN-specific ctiCorrect() in generic
//								 corrector use generic ctiCorrect()
//		UL, 2000-03-21: 1.8.0: - in medium-accuracy mode skip PN CTI correction
//								 because it has already been performed during
//								 the gain correction
//		UL, 2000-03-27: 1.9.0: - new MOS CTI correction
//		UL, 2000-04-12: 1.10.0:- downgraded error condition about CTI
//								 coefficients for MOS being out-of-bounds
//								 to noisy message
//		UL, 2000-04-25: 1.11.0:- fixed problem with RGS CTI correction (scale
//								 factor was not considered)
//		UL, 2000-05-02: 1.12.0:- fixed small problem in reading of RGS CTI
//								 coefficients
//		UL, 2000-05-04: 1.13.0:- really fix problem that was supposed to be
//								 fixed in 1.12.0
//		UL, 2000-10-05: 1.14.0:- new time-dependent CTI correction algorithm
//								 for EMOS; is backwards compatible, i.e.
//								 pre-issue 5 CTI CCFs will give the same result
//								 as with old algorithm
//		UL, 2000-11-10: 1.15.0:- new RGS CTI correction from Cor de Vries
//								 [see email dated 2000-10-10]: CTI (serial and
//								 parallel are proportional to charge)
//		UL, 2000-12-20: 1.16.0:- re-added RGS CTI correction algorithm-0; the
//								 new one (1.15.0) is now selected with ALGOID=1
//		UL, 2001-01-25: 1.17.0:- bug fix in RGS serial CTI correction:
//							     disregarded readout direction (use of chipX
//								 instead of rawX) - credit to Cor de Vries
//								 for spotting this
//		UL, 2001-04-09: 1.18.0:- made <frame> argument of EPN CTI correction
//								 of type int32
//		UL, 2001-09-14: 1.19.0:- hot/discarded pixel data for EPN CTI
//								 correction are now read from CCF::EPN_BADPIX
//								 instead of dedicated tables in CCF::EPN_CTI
//							   - EPN_CTI:CTI-HIGH can have ALGOID-1
//								 (with dropped HOT/DISCARDED_PIXEL tables)
//		UL, 2001-11-21: 1.20.0:- added HTR specific offset correction;
//								 implementation uses spectrocopy method for
//								 the time being
//		UL, 2001-12-04: 1.21.0:- altered *CTI*:CTI-EXTENDED reading to make
//								 it compliant with emos-cti-0.7 changes
//								 (CTI_X/CTI_Y are vector columns of length 4
//								 containing all the coefficients driving
//								 new ALGOID-1 CTI correction)
//		UL, 2002-01-16: 1.21.1:- downgaded CTI mode-mismatch warning to
//								 verbose-message
//		UL, 2002-07-17: 1.22.0:- additional parameter are read from CCF
//								 and passed to pn CTI correction
//		UL, 2002-07-31: 1.23.0:- added ALGOID=2 MOS CTI correction
//	   RDS, 2006-05-29: 1.24.0:- added code for MOS column dependent correction
//
//	AUTHOR:
//		U. Lammers, Thu Jun 17 15:21:51 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.
//		[2]: Calibration Access and Data Handbook, C. Erd et. al.
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/CtiCorrector.cc,v 1.63 2008/06/23 13:50:45 rsaxton Exp $
//

#include <algorithm>	//	count_if
#include <cmath>
#include <iomanip>

#include "Tracer.h"
#include "errstr.h"
#include "common.h"
#include "Xmm.h"
#include "CalOalUtility.h"
#include "CtiCorrector.h"
#include "CalPnAlgo.h"
#include "CalServer.h"
#include "GeometryDataServer.h"
#include "TimeCorrector.h"
#include "Image.h"

using namespace std;

//
//	---( CtiCorrector )---------------------------------------------------
//
CalAtomCollection<CtiCorrector>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;

	_nCCDs	= instrument->nCcds();
	_nNodes	= instrument->nNodes();
	_nModes	= instrument->nModes();
	_nEntries	= _nCCDs * _nModes;
	CtiCorrector *null = 0;
	_collection = new vector<CtiCorrector *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::CTI);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet, "CTI_EXTENDED", 2);
	const Table *table = ccftab.block();
	ColumnData *cdccdId = table->column("CCD_ID")->data(),
			   *cdnodeId = table->column("NODE_ID")->data(),
			   *cdmodeId = table->column("MODE_ID")->data();

	int8	*ccdId	= cdccdId->int8Data(),
			*nodeId = cdnodeId->int8Data(),
			*modeId = cdmodeId->int8Data();

   // Define a vector for CCDs to give value for each corrector element
    vector<int8> ccdOut(_nEntries); 

	unsigned N = table->rows();
	for (unsigned n=0; n<N; ++n) {
		unsigned pos = indexCcdMode(ccdId[n], modeId[n]);

       // Set the CCD value to be used later
        ccdOut[pos]=ccdId[n];

		if ((*_collection)[pos])
			continue;

		if (ccdId[n] <= 0 || ccdId[n] > INT(_nCCDs) ||
			nodeId[n] >= INT(_nNodes) ||
			modeId[n] >= INT(_nModes))
			errstr<<"Illegal chip and/or node/mode number in row #"
				  <<n<<" of CCF file "
				  <<Xmm::toString(instrument->id())<<"/"
				  <<Ccf::toString(Ccf::CTI)
				  <<Sas::error("IllegalValues");

		size_t nNodes = _nNodes;
		vector<vector<double> > parallelC(nNodes), serialC(nNodes);
		vector<double> coef;
		for (unsigned m=n; m<N; ++m)
			if (ccdId[n] == ccdId[m] && modeId[n] == modeId[m]) {
				//	read parallel coefficients
				CalOalUtility::fillVector(table, "CTI_Y", m, coef);
				parallelC[nodeId[m]] = coef;
				CalOalUtility::fillVector(table, "CTI_X", m, coef);
				serialC[nodeId[m]] = coef;
			}

		//
		//	construct new CtiCorrector object
		//
		(*_collection)[pos] = new CtiCorrector(parallelC, serialC, this);
	}

	//	tidy up
	delete cdccdId;
	delete cdnodeId;
	delete cdmodeId;

    /*  Declare and initialise some pointers which are needed for
        MOS column-dependent correction (not permitted inside 'case') */
    int8 *ccdId_col=0, *nodeId_col=0;
    int16 *rawx=0, *rawy_start=0, *ylength=0;
    real32  *offset=0;

    ColumnData *ctCcdId=0, *ctNodeId=0, *ctRawX=0, *ctRawYst=0,
               *ctYlength=0, *ctoffset=0;

    /*  Set up images to hold the offset at each pixel for each node 
        for the MOS column-dependent correction */
    int maxx = static_cast<int16>(instrument->maxX()),
        maxy = static_cast<int16>(instrument->maxY());          
    SImage map0(maxx+1, maxy+1),
           map1(maxx+1, maxy+1);        
    int colrows=0;

	//
	//	instantiate high-accuracy/instrument-specific Cti corrector
	//
	switch (instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:

      /* Read in the MOS column-dependent CTI data if it exists */
          if (dataSet->hasBlock("CTI_COLUMN"))
          {
            CcfConstituentTable colblk(dataSet, "CTI_COLUMN", 2);
            const Table *coltab = colblk.block();
            colrows = coltab->rows();
            ctCcdId = coltab->column("CCD_ID")->data(),
            ctNodeId = coltab->column("NODE_ID")->data(),
            ctRawX = coltab->column("RAWX")->data(),
            ctRawYst = coltab->column("RAWY_START")->data(),
            ctYlength = coltab->column("YLENGTH")->data(),
            ctoffset = coltab->column("OFFSET")->data();
                                                                                
            ccdId_col  = ctCcdId->int8Data();
            nodeId_col = ctNodeId->int8Data();
            rawx = ctRawX->int16Data();
            rawy_start = ctRawYst->int16Data();
            ylength = ctYlength->int16Data();
            offset = ctoffset->real32Data();
          }

      /* Set up a corrector for each CCD, observing mode combination */
		  for (unsigned n=0; n<_nEntries; ++n) {

			if ((*_collection)[n]) {

                vector<SImage> colOffsets; 

              /* zero the maps */
                for (int ix=0; ix<=maxx; ix++) {
                 for (int iy=0; iy<=maxy; iy++) {
                     map0[ix][iy]=0.0;
                     map1[ix][iy]=0.0;
                 }
                }

              // Loop over entries in the cti_column table if existing
                if (dataSet->hasBlock("CTI_COLUMN")) {

                  for (int r=0; r<colrows; r++) {
                                                                                
                       // cout<<"Starting RDS "<<ccdId_col[r]*1<<" "<<r
                         //   <<" "<<rawx[r]<<" "<<ccdOut[n]*1<<endl;

                   // Add to the offset maps if correct CCD
                      if (ccdId_col[r] == ccdOut[n])
                      {
                       // Set each element in the map for this node
                        //cout<<"Starting RDS "<<rawx[r]<<" "<<r<<" "<<colrows
                          //  <<rawy_start[r]<<endl;

                        for (int16 iy=0; iy<ylength[r]; iy++) {
                           if (nodeId_col[r] == 0) {
                              map0[rawx[r]][rawy_start[r]+iy]=offset[r];
                           } else {
                              map1[rawx[r]][rawy_start[r]+iy]=offset[r];
                           }
                        } // Loop over ylength pixels
                      //  cout<<"Finishing RDS"<<endl;

                      } // If this CCD wanted
                  } // Loop over cti column CCF entries
                } // If column CTI block exists
             
             // Write the offset maps for each node into a vector
                colOffsets.push_back(map0);
                colOffsets.push_back(map1);

             // Set up the corrector for this CCD, mode combination
			    CtiCorrector *p = (*_collection)[n];
				(*_collection)[n] = new EmosCtiCorrector(*p, table,
										ccftab.algoid(), colOffsets);
				delete p;
		    }
		  }

          // Remove column pointers
          if (dataSet->hasBlock("CTI_COLUMN")) {
            delete ctCcdId;
            delete ctNodeId;
            delete ctRawX;
            delete ctRawYst;
            delete ctYlength;
            delete ctoffset;            
          }

		  break;
		case Xmm::EPN:
			if (accuracy != LOW) {

				for (unsigned n=0; n<_nEntries; ++n) {
					if ((*_collection)[n]) {
						CtiCorrector *p = (*_collection)[n];
						(*_collection)[n] = new EpnCtiCorrector(*p);
						delete p;
					}
				}
				//
				//	read the Cti parameters
				//
				EpnCtiCorrector::setCtiParams(dataSet, e);
			}
			break;
		case Xmm::RGS1:
		case Xmm::RGS2: {
			CcfConstituentTable ccftab(dataSet, "XCTI", 1); // we know about
															// ALGOID: 0+1
			const Table *xctiTab = ccftab.block();
			double scaleX = xctiTab->attribute("SCALE")->asReal();
			ColumnData *cdCcd = xctiTab->column("CCD_ID")->data(),
					   *cdNode = xctiTab->column("NODE_ID")->data(),
					   *cdCtiX = xctiTab->column("CTI_X")->data();
			int8 *ccdid = cdCcd->int8Data(),
				 *nodeid = cdNode->int8Data();
			real32 *ctiX = cdCtiX->real32Data();
			for (unsigned ccd=1; ccd<=_nCCDs; ++ccd) {
				vector<double> xCtiVec(_nNodes);
				vector<bool> present(_nNodes);
				for (unsigned n=0; n<_nNodes; ++n) {
					xCtiVec[n] = 0.;
					present[n] = false;
				}
				for (unsigned n=0; n<xctiTab->rows(); ++n)
					if (ccdid[n] == ccd) {
						xCtiVec[nodeid[n]] = scaleX * ctiX[n];
						present[nodeid[n]] = true;
					}
				for (unsigned n=0; n<_nNodes; ++n)
					if (!present[n])
						errstr<<"No entry for CCD/node "<<ccd<<'/'<<n
							  <<" in `"<<xctiTab->qualifiedName()
							  <<Sas::error("IncompleteSerialCtiData");

				CcfConstituentTable ccftab(dataSet, "CTIY" + toString(ccd), 1);
				const Table *ccdTab = ccftab.block();
				double scaleY = ccdTab->attribute("SCALE")->asReal();
				vector<double>	ycol, ctiY;
				ColumnData *cdycol = ccdTab->column("YCOL")->data(),
						   *cdctiY = ccdTab->column("CTI_Y")->data();
				int16	   *ycolDat = cdycol->int16Data();
				real32	   *ctiYdat = cdctiY->real32Data();

				unsigned N = ccdTab->rows();
				for (unsigned n=0; n<N; ++n) {
					ycol.push_back(DBL(ycolDat[n]));
					ctiY.push_back(scaleY * DBL(ctiYdat[n]));
					SAS_DEBUG(cout<<ycol[n]<<' '<<ctiY[n]<<endl;)
				}

				delete cdycol;
				delete cdctiY;

				for (unsigned mode=0; mode<_nModes; ++mode) {
					unsigned pos = indexCcdMode(ccd, mode);
					if ((*_collection)[pos]) {
						CtiCorrector *p = (*_collection)[pos];
						(*_collection)[pos] =
										new RgsCtiCorrector(*p, xCtiVec,
															ycol, ctiY,
															ccftab.algoid());
						delete p;
					}
				}
			}
			delete cdCcd;
			delete cdNode;
			delete cdCtiX;
			}
			break;
		default:
			break;
//				errstr<<"Medium/high CTI correction algorithm for "
//						"instrument "<<instrument->id()<<" is not supported - "
//						"will use generic one"
//					  <<Sas::warning("UnsupportedCtiCorrection");
	}

	dataSetServer->close(dataSet);
} // CalAtomCollection<CtiCorrector>::CalAtomCollection


CtiCorrector *CalAtomCollection<CtiCorrector>::select(
					Ccd ccd,
					Node node,
					unsigned modeOffset) const
{
	_ccd = ccd;
	_node = node;
	unsigned n = indexCcdMode(ccd, modeOffset);

	if (!(*_collection)[n]) {
		errstr<<"The CCF does not contain CTI data for mode `"
			  <<calServer.state()->mode()->get()
			  <<"\' - retrieving data for "
				"fallback mode `"
			  <<calServer.instrument()->defaultMode()<<"'"
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		n = indexCcdMode(ccd, 0);
		assert((*_collection)[n]);
	}

	return (*_collection)[n];
} // CalAtomCollection<CtiData>::select()



void CtiCorrector::dump(ostream &os)
{
	os<<"CtiCorrector{\n";
	os<<"parallel coefficients: ";
	for (unsigned n=0; n<_parallelCoefficients.size(); ++n) {
		os<<"node: "<<n<<endl;
		for (unsigned k=0; k<_parallelCoefficients[n].size(); ++k)
			os<<_parallelCoefficients[n][k]<<" ";
		cout<<endl;
	}
	for (unsigned n=0; n<_serialCoefficients.size(); ++n) {
		os<<"node: "<<n<<endl;
		for (unsigned k=0; k<_serialCoefficients[n].size(); ++k)
			os<<_serialCoefficients[n][k]<<" ";
		cout<<endl;
	}

	os<<"}";
} // CtiCorrector::dump()


static inline double cti(double lnctex, double lnctey, int16 rawX, int16 rawY,
						 real32 eIn)
{
	return eIn/exp(DBL(rawX) * lnctex)/exp(DBL(rawY) * lnctey);
} // cti()

CalReal32Vector &CtiCorrector::ctiCorrect(const CalReal64Vector &,
										  const CalInt16Vector &rawX,
										  const CalInt16Vector &rawY,
										  const CalReal32Vector &eIn,
										  CalReal32Vector &eOut,
										  CalInt8Vector *nodes) const
{
	errstr<<"Time dependence of CTI correction currently not implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	int nNodes = _serialCoefficients.size();
	double lnctex = 0.,
		   lnctey = 0.;
	unsigned N = eIn.size();
	if (!nodes) {
		int nodeIndex = parent()->node();
		if (nodeIndex < nNodes) {
			lnctex = log(1.-_serialCoefficients[nodeIndex][0]);
			lnctey = log(1.-_parallelCoefficients[nodeIndex][0]);
		} else
			errstr<<"No offset CTI coefficients for node "<<nodeIndex
				  <<Sas::error("NoSuchValue");

	    for (unsigned i=0; i<N; ++i)
			eOut[i] = cti(lnctex, lnctey, rawX[i], rawY[i], eIn[i]);
	} else {
		CalInt8Vector &nodesVec = *nodes;
		for (unsigned i=0; i<N; ++i) {
			if (nodesVec[i] < nNodes) {
				lnctex = log(1.-_serialCoefficients[nodesVec[i]][0]);
				lnctey = log(1.-_parallelCoefficients[nodesVec[i]][0]);
			} else {
				errstr<<"No CTI coefficients for node "<<INT(nodesVec[i])
					  <<" event #"<<i<<" - no correction applied"
					  <<Sas::warning("NoSuchValue");
				lnctex = lnctey = 0.;
			}

			eOut[i] = cti(lnctex, lnctey, rawX[i], rawY[i], eIn[i]);
		}
	}

	return eOut;
} // CtiCorrector::ctiCorrect()


CalReal32Vector &CtiCorrector::ctiCorrect(const CalReal64Vector &time,
										  const CalInt16Vector &rawX,
										  const CalInt16Vector &rawY,
										  const CalInt32Vector &,	 // frame
										  const CalReal32Vector &eIn,
										  CalReal32Vector &eOut) const
{
	unsigned N = rawX.size();
	CalInt8Vector nodeVec(0, N);
	for (unsigned n=0; n<N; ++n)
		nodeVec[n] = rawX[n]-1;

	this->ctiCorrect(time, rawX, rawY, eIn, eOut, &nodeVec);

	return eOut;
} // CtiCorrector::ctiCorrect()


//
//	EMOS-specific CTI correction
//
EmosCtiCorrector::EmosCtiCorrector(const CtiCorrector &genericCtiCorrector,
								   const Table *ctiTable,
								   unsigned short algoid,
                                   vector<SImage> colOffsets)
	: CalChild<CalAtomCollection<CtiCorrector> >(genericCtiCorrector.parent()),
	  CtiCorrector(genericCtiCorrector),
	  _refDate(Xmm::missionReferenceTime),
	  _serialCtiRate(0), _parallelCtiRate(0), _colOffsets(colOffsets)
{
	if (ctiTable->hasAttribute("REF_DATE")) {
		_refDate = STime(ctiTable->attribute("REF_DATE")->asString());
	}

	if (ctiTable->hasAttribute("CTX_RATE")) {
		_serialCtiRate = ctiTable->attribute("CTX_RATE")->asReal();
		_parallelCtiRate = ctiTable->attribute("CTY_RATE")->asReal();
	}

	_algoid = algoid;
} // EmosCtiCorrector::EmosCtiCorrector()


CalReal32Vector &EmosCtiCorrector::ctiCorrect(const CalReal64Vector &,
											  const CalInt16Vector &rawX,
											  const CalInt16Vector &rawY,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut,
											  CalInt8Vector *nodes) const
{
	errstr<<"Time dependence of CTI correction currently not implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	double deltaT = calServer.state()->exposureStartTime()->get() -
					_refDate;

	int nNodes = _serialCoefficients.size();
	unsigned N = eIn.size();
	int nodeIndex = parent()->node();

	switch (_algoid) {
		case 0:
			//	ALGOID=0 algorithm
			for (unsigned n=0; n<N; ++n) {
				if (nodes)
					nodeIndex = (*nodes)[n];
				if (nodeIndex >= nNodes) {
					errstr<<"No CTI coefficients for node "<<nodeIndex
						  <<" event #"<<n<<" - no correction applied"
						  <<Sas::warning("NoSuchValue");
					eOut[n] = eIn[n];
					continue;
				}

				//
				//	email DL 2000-10-05
				//
				double pLossPerTransfer =
							deltaT * _parallelCtiRate * sqrt(eIn[n]) +
								_parallelCoefficients[nodeIndex][0] +
								_parallelCoefficients[nodeIndex][1] * eIn[n],
					   sLossPerTransfer =
							deltaT * _serialCtiRate +
								_serialCoefficients[nodeIndex][0] +
								_serialCoefficients[nodeIndex][1] * eIn[n];

				eOut[n] = eIn[n] +
					rawY[n] * pLossPerTransfer + rawX[n] * sLossPerTransfer;
			}
			break;
		case 1:
			//	ALGOID=1 algorithm
			for (int n=0; n<nNodes; ++n)
				if (_serialCoefficients[n].size() != 4 ||
					_parallelCoefficients[n].size() != 4)
					errstr<<"EMOS CTI coefficient vectors have wrong sizes - "
							"expected 4 but got "
						  <<_serialCoefficients[n].size()<<'/'
						  <<_parallelCoefficients[n].size()
						  <<Sas::error("CcfDataIntegrityError");

			for (unsigned n=0; n<N; ++n) {
				if (nodes)
					nodeIndex = (*nodes)[n];
				if (nodeIndex >= nNodes) {
					errstr<<"No CTI coefficients for node "<<nodeIndex
						  <<" event #"<<n<<" - no correction applied"
						  <<Sas::warning("NoSuchValue");
					eOut[n] = eIn[n];
					continue;
				}

				//
				//	email DL 2001-09-24
				//
				const vector<double> &serial = _serialCoefficients[nodeIndex],
									 &parall =_parallelCoefficients[nodeIndex];
				double ctiX = serial[0] + serial[1] * eIn[n] +
								  deltaT * serial[2] *
								  std::pow(static_cast<double>(eIn[n]),
										   serial[3]),
					   ctiY = parall[0] + parall[1] * eIn[n] +
								  deltaT * parall[2] *
								  std::pow(static_cast<double>(eIn[n]),
										   parall[3]);

				eOut[n] = eIn[n] + rawY[n]*ctiY + rawX[n]*ctiX;
			}
			break;
		case 2:
			//	ALGOID=2 algorithm
			for (int n=0; n<nNodes; ++n)
				if (_serialCoefficients[n].size() != 3 ||
					_parallelCoefficients[n].size() != 3)
					errstr<<"EMOS CTI coefficient vectors have wrong sizes - "
							"expected 3 but got "
						  <<_serialCoefficients[n].size()<<'/'
						  <<_parallelCoefficients[n].size()
						  <<Sas::error("CcfDataIntegrityError");

			for (unsigned n=0; n<N; ++n) {
				if (nodes)
					nodeIndex = (*nodes)[n];
				if (nodeIndex >= nNodes) {
					errstr<<"No CTI coefficients for node "<<nodeIndex
						  <<" event #"<<n<<" - no correction applied"
						  <<Sas::warning("NoSuchValue");
					eOut[n] = eIn[n];
					continue;
				}

				//
				//	email BA 2002-07-31
				//
				const vector<double> &serial = _serialCoefficients[nodeIndex],
									 &parall =_parallelCoefficients[nodeIndex];
				double ctiX = (serial[0] + deltaT * serial[1]) *
								  std::pow(static_cast<double>(eIn[n]),
										   serial[2]),
					   ctiY = (parall[0] + deltaT * parall[1]) *
								  std::pow(static_cast<double>(eIn[n]),
										   parall[2]);

              /* The CTI is calculated as a fn of the raw coordinates. 
                 The special case of the FAST modes must be handled in the
                 calling application */
				eOut[n] = eIn[n] + rawY[n]*ctiY + rawX[n]*ctiX;

              /* Add on the column specific offset, if pixel within
                 normal CCD bounds */
                int16 xExt=_colOffsets[nodeIndex].xExtent(),
                      yExt=_colOffsets[nodeIndex].yExtent();

                if ( rawX[n]<xExt && rawY[n]<yExt )
                    eOut[n] -= _colOffsets[nodeIndex][rawX[n]][rawY[n]];
			}
			break;
		default:
			errstr<<"Unsupported ALGOID "<<_algoid
				  <<Sas::error("UnsupportedAlgoId");
			break;
	}

	return eOut;
} // EmosCtiCorrector::ctiCorrect()


EmosCtiCorrector::~EmosCtiCorrector()
{
} // EmosCtiCorrector::~EmosCtiCorrector()



//
//	EPN-specific CTI correction
//

//	data members
bool EpnCtiCorrector::_ctiParamsSet = false;
EpnCtiCorrector::CtiParam EpnCtiCorrector::_ctiParam;
EpnCtiCorrector::HotPixelParam EpnCtiCorrector::_hotPixelParam;
EpnCtiCorrector::DiscardedPixelParam EpnCtiCorrector::_discardedPixelParam;
double EpnCtiCorrector::_ePerAdu = 1.,
	   EpnCtiCorrector::_keVPerE = 1.;
vector<unsigned short> EpnCtiCorrector::_yShift;
vector<float> EpnCtiCorrector::_ctiScale,
			  EpnCtiCorrector::_eFFpar,
			  EpnCtiCorrector::_modeFactor,
			  EpnCtiCorrector::_swPar,
			  EpnCtiCorrector::_lwPar;

void EpnCtiCorrector::setCtiParams(const DataSet *set, const CifEntry *e)
{
	CcfConstituentTable ccftab(set, e->queryBlockName("CTI-HIGH"), 1);
	const Table *table = ccftab.block();
	_ePerAdu = table->attribute("EPERADU")->asReal();	
	_keVPerE = table->attribute("KEVPER_E")->asReal();
	if (table->rows() != N_CTI_PARAMETERS)
		errstr<<"Expected to find "<<N_CTI_PARAMETERS<<" CTI parameters "
				"in PN CTI table: "
			  <<table->qualifiedName()<<" but there are only "
			  <<table->rows()
			  <<Sas::error("DataMismatch");

	const Column *ctiCol = table->column("CTI_PAR");
	ColumnData *ctiParCd = ctiCol->data();
	unsigned nElements = ctiParCd->elements();
	if (nElements != xmmPayLoad.epn.nNodes() * xmmPayLoad.epn.nCcds())
		errstr<<"Expected to find "<<xmmPayLoad.epn.nNodes() *
									 xmmPayLoad.epn.nCcds()
			  <<" elements in column "<<ctiCol->qualifiedName()
			  <<" but there are "<<nElements
			  <<Sas::error("DataMismatch");

	real32 *ctiPar = ctiParCd->real32Data();

	//
	//	allocate memory and read data
	//

	//	CTI parameters
	_ctiParam.n = nElements;
	_ctiParam.iflParam = new int [nElements];
	for (unsigned n=0; n<N_CTI_PARAMETERS-1; ++n)
		_ctiParam.ctiParam.push_back(new float [nElements]);

	for (unsigned pos=0; pos<nElements; ++pos)
		_ctiParam.iflParam[pos] = INT(ctiPar[ctiParCd->position(0, pos)]);
	for (unsigned n=0; n<N_CTI_PARAMETERS-1; ++n)
		for (unsigned pos=0; pos<nElements; ++pos)
			_ctiParam.ctiParam[n][pos] =
									ctiPar[ctiParCd->position(n+1, pos)];
	delete ctiParCd;

	//
	//	prepare hot/discarded pixels data for CTI correction; hot pixels
	//	are not-uplinked bad pixels of type 'H' (hot) in EPN_BADPIX_xxxx.CCF;
	//	discarded pixels are uplinked hot pixels (type 'h') which are
	//	not telemetered by the event analyzer logic, i.e. appear as logically
	//	dead
	//
	BadPixelDataServer *bpds = 0; bpds = calServer.getAtom(bpds);
	BadPixelMap badPixelMap = bpds->map();
	BadPixelMap::const_iterator i, begin = badPixelMap.begin(),
								   end = badPixelMap.end();
	_hotPixelParam.n = std::count_if(begin, end,
									 isDefectAndNotUplinked<BadPixel::HOT>());
	_discardedPixelParam.n = std::count_if(begin, end,
									 isDefectAndUplinked<BadPixel::HOT>());

	//
	//	allocate space for hot/discarded pixel data
	//
	for (unsigned n=0; n<5; ++n)
		_hotPixelParam.hotPixelParam.push_back(new int [_hotPixelParam.n]);
	_hotPixelParam.saturatedImp = new float [_hotPixelParam.n];

	for (unsigned n=0; n<4; ++n)
		_discardedPixelParam.discardedPixelParam.push_back(
										new int [_discardedPixelParam.n]);


	unsigned nHot = 0,
			 nDisc = 0;
	Ccd ccd = calServer.state()->ccd()->get();
	int quadId = EPNcam::toQuadrantId(ccd),	//	HW quadrant number [0, 3]
		ccdId = EPNcam::toCcdId(ccd);		//	HW CCD number [0, 2]
	for (i=begin; i!=end; ++i) {
		if (i->second.type() == BadPixel::HOT) {
			const BadPixel &bp = i->second;
			int rawX = i->first.first-1,	//	HW rawX coordinate [0, 63]
				rawY = i->first.second -1;	//	HW rawY coordinate [0, 199]
			if (!bp.uplinked()) {
				//	hot pixel
				_hotPixelParam.hotPixelParam[0][nHot] = quadId;
				_hotPixelParam.hotPixelParam[1][nHot] = ccdId;

				//	HW RAWX(0..63)/RAWY(0..199)
				_hotPixelParam.hotPixelParam[2][nHot] = rawX;
				_hotPixelParam.hotPixelParam[3][nHot] = rawY;

				//	the number of pixels that are affected by this hottie
				//	set to constant 2 for the time being awaiting formula
				//	from MPE (email MJF 2001-09-13) to compute from
				//	number of saturated impurities
				_hotPixelParam.hotPixelParam[4][nHot] = 2;

				//	number of saturated impurities
				_hotPixelParam.saturatedImp[nHot] =	bp.auxAttribute();
				++nHot;
			} else {
				//	discarded pixel
				_discardedPixelParam.discardedPixelParam[0][nDisc] = quadId;
				_discardedPixelParam.discardedPixelParam[1][nDisc] = ccdId;
				_discardedPixelParam.discardedPixelParam[2][nDisc] = rawX;
				_discardedPixelParam.discardedPixelParam[3][nDisc] = rawY;
				++nDisc;
			}
		}
	}

	//
	//	additional CTI parameters
	//
	table = set->table("CTI-HIGH_ADD_PAR");
	CalOalUtility::fillVector(table, "Y_SHIFT", 0, _yShift);
	assert(_yShift.size() == 6);
	CalOalUtility::fillVector(table, "SCALE", 0, _ctiScale);
	assert(_yShift.size() == 6);
	CalOalUtility::fillVector(table, "EFF_PAR", 0, _eFFpar);
	assert(_eFFpar.size() == 4);
	CalOalUtility::fillVector(table, "MODE_FACTOR", 0, _modeFactor);
	assert(_modeFactor.size() == 6);

    // Read the SW and LW parameters - keep CCF backwards compatibility 
    // by taking the old LWSW_PAR array if the newer SW_PAR array not present.
    if (table->hasColumn("SW_PAR")) {
	   errstr<<"Reading from new array"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	   CalOalUtility::fillVector(table, "SW_PAR", 0, _swPar);
	   CalOalUtility::fillVector(table, "LW_PAR", 0, _lwPar);
    } else { // Old-style CCF 
	   CalOalUtility::fillVector(table, "LWSW_PAR", 0, _swPar);
	   CalOalUtility::fillVector(table, "LWSW_PAR", 0, _lwPar);
    }
	assert(_swPar.size() == 16);

	_ctiParamsSet = true;
} // EpnCtiCorrector::setCtiParams()


//
//	the low-accuracy CTI correction
//
CalReal32Vector &EpnCtiCorrector::ctiCorrect(const CalReal64Vector &,
											 const CalInt16Vector &,
											 const CalInt16Vector &rawY,
											 const CalReal32Vector &eIn,
											 CalReal32Vector &eOut,
											 CalInt8Vector *) const
{
	errstr<<"Time dependence of CTI correction currently not implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	double lnctey = log(1.-parallelCoefficients()[0][0]);
	unsigned N = eIn.size();
    for (unsigned i=0; i<N; ++i)
		eOut[i] = eIn[i]/exp(DBL(rawY[i]) * lnctey);

	return eOut;
} // EpnCtiCorrector::ctiCorrect()


//
//	high-accuracy EPN CTI correction
//
CalReal32Vector &EpnCtiCorrector::ctiCorrect(const CalReal64Vector &, // time
											 const CalInt16Vector &rawX,
											 const CalInt16Vector &rawY,
											 const CalInt32Vector &frame,
											 const CalReal32Vector &eIn,
											 CalReal32Vector &eOut) const
{
	if (calServer.state()->accuracyLevel()->get() == MEDIUM) {
		//
		//	in medium-accuracy mode the energy has already been CTI corrected
		//	during the gain correction
		//
		unsigned N = eIn.size();
		for (unsigned n=0; n<N; ++n)
			eOut[n] = eIn[n];

		return eOut;
	}

	//	first initialize the PN CTE corrector  algorithm

	//	CTE parameters
	unsigned NPARAM = _ctiParam.n;
	int *iflparam = _ctiParam.iflParam;
	float *a = _ctiParam.ctiParam[0],
		  *b = _ctiParam.ctiParam[1],
		  *c = _ctiParam.ctiParam[2];

	//
	//	get CCD mode and temperature from the state
	//
	float temp = calServer.state()->temperature()->get();	// [K]
	if (calServer.state()->temperatureTracking()->get())
		errstr<<"Automatic temperature tracking is requested but this "
				"is not yet implemented - taking CCD temperature from "
				"state variable `temperature' which has the value "
			  <<temp<<" K."<<Sas::warning("TemperatureTracking");
	unsigned modeId = calServer.instrument()->modeIndex(
											calServer.state()->mode()->get());
	float ytiming = calServer.state()->auxiliaryParameter()->get();

	//	obtain frame time + time correction parameter
	TimeCorrector *tc = 0; tc = calServer.getAtom(tc);
	float frameTime = tc->frameTime()/1000.,	//	ms->s
		  timePar = tc->correctionParameters()
				[2][calServer.instrument()->modeIndex(Xmm::FAST_TIMING)]/1000.;

	assert(_yShift.size() > modeId);
	int	  yShift = _yShift[modeId];
	assert(_ctiScale.size() > modeId);
	float ctiScale	 = _ctiScale[modeId];

	float eFFparin[] = { _eFFpar[0], _eFFpar[1], _eFFpar[2], _eFFpar[3] },
		  keVPerein = _keVPerE;

	float modeFactor = _modeFactor[modeId];

    // Set the LW or SW array depending on the observing mode
	float lwswPar[_swPar.size()];
	for (unsigned n=0; n<_swPar.size(); ++n)
       if (calServer.state()->mode()->get() == Xmm::PRIME_SMALL_WINDOW) {
	    	lwswPar[n] = _swPar[n];
       } else {
            lwswPar[n] = _lwPar[n];
       }

	CALPNALGO_initCteParam(NPARAM, iflparam, a, b, c, temp, modeId, ytiming,
						   frameTime, timePar, yShift, ctiScale, eFFparin,
						   keVPerein, modeFactor, lwswPar);

	//	hot pixel list
	unsigned nHotPixels = _hotPixelParam.n;
	int *hotquadrant = _hotPixelParam.hotPixelParam[0],
		*hotccd = _hotPixelParam.hotPixelParam[1],
		*hotrawX = _hotPixelParam.hotPixelParam[2],
		*hotrawY = _hotPixelParam.hotPixelParam[3],
		*hotyLen = _hotPixelParam.hotPixelParam[4];
		float *hotsaturatedImp = _hotPixelParam.saturatedImp;

	CALPNALGO_initHotPixels(nHotPixels, hotquadrant, hotccd, hotrawX, hotrawY,
							hotsaturatedImp, hotyLen);

	//	discarded pixel list
	unsigned nDiscardedPixels = _discardedPixelParam.n;
	int *discquadrant = _discardedPixelParam.discardedPixelParam[0],
		*discccd = _discardedPixelParam.discardedPixelParam[1],
		*discrawX = _discardedPixelParam.discardedPixelParam[2],
		*discrawY = _discardedPixelParam.discardedPixelParam[3];

	CALPNALGO_initDiscardedPixels(nDiscardedPixels, discquadrant, discccd,
								  discrawX, discrawY);

	//
	//	perform the correction
	//
	unsigned N = eIn.size();
	Ccd	ccd = parent()->ccd();
	unsigned quadId = EPNcam::toQuadrantId(ccd),
			 ccdId = EPNcam::toCcdId(ccd);

	for (unsigned n=0; n<N; ++n) {
		int		 frameNo = frame[n];
		unsigned quadCol = EPNcam::toQuadrantColumnNumber(quadId, ccdId,
														  rawX[n]),
				 quadRow = rawY[n] - 1;
		int		 ierr = 0;
		float	 uncorrectedAmp = eIn[n] * _ePerAdu,
				 correctedAmp;

		CALPNALGO_ctiCorrect(frameNo, quadId, quadCol, quadRow, uncorrectedAmp,
							 correctedAmp, ierr);

		if (!ierr)
			eOut[n] = correctedAmp / _ePerAdu;
		else {
			errstr<<"CTI correction algorithm for event #"<<n
				  <<" (CCD/rawX/rawY = "<<ccd<<"["<<quadId<<"/"<<ccdId<<"]/"
				  <<rawX[n]<<"/"<<rawY[n]<<"["<<quadCol<<"/"<<quadRow
				  <<"]) failed - error code: "<<ierr
				  <<" - no correction applied - reason for failure: ";
			switch (ierr) {
				case 1: errstr<<"wrong pixel coordinates"; break;
				case 2: errstr<<"very bright pixel/to be discarded"; break;
				case 3: errstr<<"the column has no calibration data"; break;
				default: errstr<<"unknown"; break;
			}
			errstr<<Sas::warning("ctiCorrectionFailed");
			eOut[n] = eIn[n];
		}
	}

	return eOut;
} // EpnCtiCorrector::ctiCorrect()


EpnCtiCorrector::~EpnCtiCorrector()
{
	//	free up memory
	for (unsigned n=0; n<_ctiParam.ctiParam.size(); ++n) {
		delete [] _ctiParam.ctiParam[n]; _ctiParam.ctiParam[n] = 0;
	}
	delete [] _ctiParam.iflParam; _ctiParam.iflParam = 0;

	for (unsigned n=0; n<_hotPixelParam.hotPixelParam.size(); ++n) {
		delete [] _hotPixelParam.hotPixelParam[n];
		_hotPixelParam.hotPixelParam[n] = 0;
	}
	delete [] _hotPixelParam.saturatedImp; _hotPixelParam.saturatedImp = 0;

	for (unsigned n=0; n<_discardedPixelParam.discardedPixelParam.size(); ++n){
		delete [] _discardedPixelParam.discardedPixelParam[n];
		_discardedPixelParam.discardedPixelParam[n] = 0;
	}
} // EpnCtiCorrector::~EpnCtiCorrector()


//
//	---( RgsCtiCorrector )--------------------------------------------------
//
RgsCtiCorrector::RgsCtiCorrector(const CtiCorrector &genericCtiCorrector,
								 const vector<double> &ctiX,
								 const vector<double> &xCol,
								 const vector<double> &ctiY,
								 unsigned short algoid)
	: CalChild<CalAtomCollection<CtiCorrector> >(genericCtiCorrector.parent()),
	  CtiCorrector(genericCtiCorrector),
	  _xCti(ctiX),
	  _yCti(xCol, ctiY)
{
	_algoid = algoid;
} // RgsCtiCorrector::RgsCtiCorrector()


CalReal32Vector &RgsCtiCorrector::ctiCorrect(const CalReal64Vector &time,
											 const CalInt16Vector &rawX,
											 const CalInt16Vector &rawY,
											 const CalInt8Vector &shape,
											 const CalInt16Vector &separation,
											 const CalReal32Vector &eIn,
											 CalReal32Vector &eOut,
											 CalInt8Vector *nodes) const
{
	switch (_algoid) {
		case 0: return ctiCorrect0(time, rawX, rawY, shape, separation,
									   eIn, eOut, nodes);
				break;
		case 1: return ctiCorrect1(time, rawX, rawY, shape, separation,
									   eIn, eOut, nodes);
				break;
		default:
				errstr<<"Unsupported ALGOID "<<_algoid<<" encountered - this "
						"should not happen."<<Sas::error("UnsupportedAlgoId");
				break;
	}

	return eOut;
} // RgsCtiCorrector::ctiCorrect()


//
//	CTI correction in HTR mode
//
CalReal32Vector &RgsCtiCorrector::ctiCorrect(const CalReal64Vector &time,
											 const CalInt16Vector &rawX,
											 int16 rawY,
											 const CalReal32Vector &eIn,
											 CalReal32Vector &eOut,
											 CalInt8Vector *nodes) const
{
	CalState *state = calServer.state();

	//	verify we have HTR mode
	Mode mode = state->mode()->get();
	if (mode != Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD &&
		mode != Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD)
		errstr<<"Called RGS HTR-specific CTI correction in mode `"
			  <<mode<<"'"<<Sas::error("ImproperMode");

	//	construct dummy vectors shape/separation and call Spectroscopy
	//	CTI method
	CalInt16Vector rawYv(0, rawX.size());
	rawYv = rawY;

	CalInt8Vector shape(0, rawX.size());
	shape = 0;

	CalInt16Vector separation(0, rawX.size());
	separation = 0;

	return ctiCorrect(time, rawX, rawYv, shape, separation, eIn, eOut, nodes);
} // RgsCtiCorrector::ctiCorrect()


//
//	ALGID=0 version of RGS CTI correction - see [2] for details
//
CalReal32Vector &RgsCtiCorrector::ctiCorrect0(const CalReal64Vector &,
											  const CalInt16Vector &rawX,
											  const CalInt16Vector &rawY,
											  const CalInt8Vector &,
											  const CalInt16Vector &,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut,
											  CalInt8Vector *nodes) const
{
	errstr<<"<time> argument of RGS CTI correction currently ignored"
		  <<Sas::warning("ArgumentIgnored");
	errstr<<"<shape> argument of RGS CTI correction currently ignored"
		  <<Sas::warning("ArgumentIgnored");
	errstr<<"<separation> argument of RGS CTI correction currently ignored"
		  <<Sas::warning("ArgumentIgnored");

	//
	//	parallel CTI correction only works in CHIPCOORD for OCB=1
	//
	unsigned ocb = calServer.state()->ocb()->get();
	GeometryDataServer *cd = 0; cd = calServer.getAtom(cd);
	unsigned N = rawX.size();
	CalInt16Vector chipX(0, N), chipY(0, rawY.size());
	for (unsigned n=0; n<N; ++n) {
		chipX[n] = rawX[n];
		chipY[n] = rawY[n];
	}
	XmmPixelList pl = cd->pixCoord1List(chipX, chipY).toChipCoord(nodes);

	if (ocb != 1) {
		//	CTI model only valid for OCB = 1; convert pixel coordinates
		//	accordingly
		XmmCartesianPointList cpl = pl.toCamCoord1();

		calServer.state()->ocb()->set(1);	//	set OCB to 1
		cpl.toChipCoord(chipX, chipY);
		calServer.state()->ocb()->set(ocb);	//	reset OCB to old value
	}

	//
	//	parallel CTI correction
	//
	for (unsigned n=0; n<N; ++n)
		eOut[n] = eIn[n] + (chipY[n]-1) * _yCti(chipX[n]-1);

	SAS_DEBUG(for (unsigned n=1; n<=1024; ++n)
		cout<<n-1<<' '<<_yCti(n-1)<<endl;)

	//
	//	serial CTI correction
	//
	int8 *nodeVec = new int8[N];
	if (!nodes) {
		unsigned nodeVal = calServer.state()->node()->get();
		for (unsigned n=0; n<N; nodeVec[n++] = nodeVal);
	} else
		for (unsigned n=0; n<N; ++n)
			nodeVec[n] = (*nodes)[n];
	
	for (unsigned n=0; n<N; ++n)
		eOut[n] += FLT(rawX[n]-1) * _xCti[nodeVec[n]];

	//	tidy up
	delete [] nodeVec;

	return eOut;
} // RgsCtiCorrector::ctiCorrect()


//
//	ALGID=1 version of RGS CTI correction - see [2] for details
//
CalReal32Vector &RgsCtiCorrector::ctiCorrect1(const CalReal64Vector &,
											  const CalInt16Vector &rawX,
											  const CalInt16Vector &rawY,
											  const CalInt8Vector &,
											  const CalInt16Vector &,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut,
											  CalInt8Vector *nodes) const
{
	errstr<<"<time> argument of RGS CTI correction currently ignored"
		  <<Sas::warning("ArgumentIgnored");
	errstr<<"<shape> argument of RGS CTI correction currently ignored"
		  <<Sas::warning("ArgumentIgnored");
	errstr<<"<separation> argument of RGS CTI correction currently ignored"
		  <<Sas::warning("ArgumentIgnored");

	//
	//	parallel CTI correction only works in CHIPCOORD for OCB=1
	//
	unsigned ocb = calServer.state()->ocb()->get();
	GeometryDataServer *cd = 0; cd = calServer.getAtom(cd);
	unsigned N = rawX.size();
	CalInt16Vector chipX(0, N), chipY(0, rawY.size());
	for (unsigned n=0; n<N; ++n) {
		chipX[n] = rawX[n];
		chipY[n] = rawY[n];
	}
	XmmPixelList pl = cd->pixCoord1List(chipX, chipY).toChipCoord(nodes);
	if (ocb != 1) {
		//	CTI model only valid for OCB = 1; convert pixel coordinates
		//	accordingly
		XmmCartesianPointList cpl = pl.toCamCoord1();

		calServer.state()->ocb()->set(1);	//	set OCB to 1
		cpl.toChipCoord(chipX, chipY);
		calServer.state()->ocb()->set(ocb);	//	reset OCB to old value
	}

	//
	//	parallel CTI correction from Cor de Vries [2000-09-18]:
	//	based on Mkn 421 data, CTI correction is now proportinal to charge;
	//	set scale factor based on flux and temperature (not in model)
	//
	for (unsigned n=0; n<N; ++n)
		eOut[n] = eIn[n] * (1. + (chipY[n]-1) * _yCti(chipX[n]-1));

	SAS_DEBUG(for (unsigned n=1; n<=1024; ++n)
		cout<<n-1<<' '<<_yCti(n-1)<<endl;)

	//
	//	serial CTI correction
	//
	int8 *nodeVec = new int8[N];
	if (!nodes) {
		unsigned nodeVal = calServer.state()->node()->get();
		for (unsigned n=0; n<N; nodeVec[n++] = nodeVal);
	} else
		for (unsigned n=0; n<N; ++n)
			nodeVec[n] = (*nodes)[n];

	for (unsigned n=0; n<N; ++n)
		eOut[n] *= (1. + FLT(rawX[n]-1) * FLT(ocb) * _xCti[nodeVec[n]]);

	//	tidy up
	delete [] nodeVec;

	return eOut;
} // RgsCtiCorrector::ctiCorrect()


RgsCtiCorrector::~RgsCtiCorrector()
{
} // RgsCtiCorrector::~RgsCtiCorrector()

