//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EnergyCorrector.cc --- implementation of EnergyCorrector class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-05-25: 1.0.0: - first version (extracted from
//								 CalAtomsCommon.cc)
//		UL, 1999-06-08: 1.1.0: - minor bug fixes
//		UL, 1999-06-15: 1.2.0: - added implementation of EpnEnergyCorrector
//							     class methods
//      UL, 1999-08-13: 1.3.0: - added piTokeV()
//		UL, 1999-09-03: 1.4.0: - changed ser->grade in
//								   RgsEnergyCorrector::offsetCorrect and
//								 updated implementation
//		UL, 1999-11-24: 1.5.0: - brought in line with calpnalgo: pass
//								 additional parameter vector to gain init
//								 routine
//		UL, 2000-01-05: 1.6.0: - removed check that cross-tests pattern
//								 IDs in energy combinator coefficient
//								 columns against those in pattern lib
//								 (fixes SPR SPR 1052 : CAL_gainCorrect does
//								  not work in Timing mode for EMOS)
//		UL, 2000-02-01: 1.7.0: - made EPN-specific gainCorrect() in generic
//								 corrector use generic gainCorrect()
//		UL, 2000-02-29: 1.8.0: - brought RGS offset correction in line with
//								 [2]
//		UL, 2000-03-01: 1.9.0: - RgsEnergyCorrector::offsetCorrect: bug
//								 fix + more sensible handling of illegal
//								 event shapes
//		UL, 2000-04-04: 1.10.0:- E_RAND_MIN/E_RAND_MAX -> RAND_MIN/RAND_MAX
//		UL, 2000-04-13: 1.11.0:- check error code from PN gain correction
//		UL, 2000-05-10: 1.12.0:- overhaul in the process of implementing new
//								 MOS gain correction
//		UL, 2000-05-22: 1.13.0:- fixed problem with generic/EMOS-specific
//								 gain correction routine
//							   - fall back upon CCF if given RGS offset
//								 override table is empty
//		UL, 2000-07-31: 1.14.0:- EPN gain correction takes now real-valued
//								 input amplitude
//		UL, 2000-11-08: 1.15.0:- EmosEnergyCorrector::gainCorrect() in modes
//								 FastUncompressed/FastCompressed for CCD 1
//								 takes additional multiplicative factor
//								 (FAST_G in CCF::AduConv) into accoun
//		UL, 2000-11-14: 1.15.1:- bug fix in
//									RgsEnergyCorrector::RgsEnergyCorrector
//								 [fixes SPR 1834: Task aborts with Bus Error
//		UL, 2000-11-21: 1.15.2:- PN gain correction: better message in case
//								 of calpnalgo failure
//		UL, 2001-04-03: 1.16.0:- several changes related to ALGOID=1
//								 RGS offset correction; offset values
//								 are now time dependent:
//									offset = offsetConst + t*offsetSlope
//								 t = obsStartDate - refTime
//								 refTime comes from ADUCONV
//		UL, 2001-10-30: 1.17.0:- added implementation of new RGS HTR offset
//								 correction
//		UL, 2001-11-21: 1.18.0:- fixed problem in HTR-specific offset
//								 correction: was testing against wrong
//								 camera mode
//		UL, 2002-07-17: 1.19.0:- added three new pn-specific methods (see
//								 EnergyCorrector.h)
//		UL, 2002-11-19: 1.20.0:- implemented ALGOID=1 of EPN long-term CTE
//								 model (time-dependence modelled
//								 as 3. order polynomial; parameters mode
//								 dependent); goes with structural change
//								 in EPN_CTI CCF; all changes fully backward
//								 compatible
//		UL, 2003-02-12: 1.20.1:- fixed bug in EnergyCorrector::offsetCorrect();
//								 only surfaced with gcc-3.2
//		RDS,2004-08-04: 1.21.0:- added gain correction for PN Eff and FF mode
//
//		RDS,2008-04-23: 1.22.0:- added Rate-dependent CTI correction
//                                       for PN Fast modes
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/EnergyCorrector.cc,v 1.58 2008/04/24 10:10:53 rsaxton Exp $
//

#include <iomanip>
#include <algorithm>
#include <cmath>

#include "errstr.h"
#include "common.h"
#include "EnergyCorrector.h"
#include "CalServer.h"
#include "CalOalUtility.h"
#include "BadPixelDataServer.h"	//	INVALID_BAD_PIXEL_TABLE_CODE
#include "CalPnAlgo.h"

using namespace std;

//
//	---( EnergyCorrector )---------------------------------------------------
//
CalAtomCollection<EnergyCorrector>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	EnergyCorrector *null = 0;
	_collection = new vector<EnergyCorrector *>(_nEntries, null);

	//
	//	read the principal data from the CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::ADUConv);
	DataSet *dataSet = e->open();
	CcfConstituentTable aduConvTab(dataSet,
								   e->queryBlockName("OFFSET_GAIN"),
								   1);	//	we know about ALGOID: 0+1
															
	//
	//	instantiate high-accuracy/instrument-specific gain corrector
	//
	switch (instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2: {
			CcfConstituentTable ccftab(dataSet, "ENERGY_COMBINE");
			const Table *eCombine = ccftab.block();
			for (unsigned n=0; n<_nEntries; ++n)
				(*_collection)[n] =
					new EmosEnergyCorrector(n+1, aduConvTab, eCombine, this);
			}
			break;
		case Xmm::EPN: {
			CcfConstituentTable ccftab1(dataSet, "GAIN-HIGH"),
								ccftab2(dataSet, "REEMISSION");
			const Table *gainHigh = ccftab1.block(),
					    *reEmission = ccftab2.block();
			for (unsigned n=0; n<_nEntries; ++n)
				(*_collection)[n] =
					new EpnEnergyCorrector(n+1, aduConvTab, gainHigh,
										   reEmission, ccf, this);
			}
			break;
		case Xmm::RGS1:
		case Xmm::RGS2: {
			CifEntry *e = ccf->entry(instrument, Ccf::DarkFrame);
			DataSet *darkSet = e->open(),
					*diagSet = 0;
			CcfConstituentTable ccftab(darkSet, "DARKFRAME");
			const Table *darkFrame = ccftab.block(),
						*offsetOverride = 0;
			if (ccf->hasEntry(instrument, Ccf::DiagnosticOffset)) {
				CifEntry *e = ccf->entry(instrument, Ccf::DiagnosticOffset);
				diagSet = e->open();
				const Table *offsetTab = diagSet->table("RGSOFFSETS");
				if (offsetTab->rows())
					offsetOverride = offsetTab;
				else
					errstr<<"Specified offset override table `"
						  <<offsetTab->qualifiedName()<<"' is empty (0 rows) "
							"- offset correction will fall back upon numbers "
							"in the CCF"
						  <<Sas::warning("EmptyOffsetOverrideTable");
			}
			for (unsigned n=0; n<_nEntries; ++n) {
				(*_collection)[n] =
					new RgsEnergyCorrector(n+1, aduConvTab, darkFrame,
										   offsetOverride, this);
			}
			if (diagSet)
				dataSetServer->close(diagSet);
			dataSetServer->close(darkSet);
			}
			break;
		default:
			break;
	}

	dataSetServer->close(dataSet);
} // CalAtomCollection<EnergyCorrector>::CalAtomCollection()


EnergyCorrector *CalAtomCollection<EnergyCorrector>::select(
					Ccd ccd,
					Node node,
					unsigned) const
{
	_ccd = ccd;
	_node = node;
	unsigned n = ccd - 1;

	return (*_collection)[n];
} // CalAtomCollection<EnergyCorrector>::select


EnergyCorrector::EnergyCorrector(Ccd ccd,
								 const CcfConstituentTable &aduConvTab,
								 CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent)
{
	const Table *aduConv = aduConvTab.block();
	_bptCode = CalOalUtility::optionalAttribute(aduConv, "BPT_CODE",
						 BadPixelDataServer::INVALID_BAD_PIXEL_TABLE_CODE);

	//
    //  read PI->eV conversion values
    //
	_eVperPi = aduConv->parent()->attribute("KEV_P_PI")->asReal()*1000.;
	_eV0	 = aduConv->parent()->attribute("KEV0")->asReal()*1000.;

	ColumnData *cdccdId	= aduConv->column("CCD_ID")->data(),
			   *cdgain	= aduConv->column("GAIN")->data(),
			   *cdoffset0 = 0, *cdoffset1 = 0;

	_offsetRefTime = Xmm::missionReferenceTime;
	if (aduConvTab.algoid() == 0)
		cdoffset0 = aduConv->column("OFFSET")->data();
	else {
		cdoffset0 = aduConv->column("OFFSET_CONST")->data();
		cdoffset1 = aduConv->column("OFFSET_SLOPE")->data();
		_offsetRefTime = STime(aduConv->attribute("REFTIME")->asString());
	}

	int8	*ccdId		= cdccdId->int8Data();
	real32	*gain		= cdgain->real32Data(),
			*offset0	= cdoffset0->real32Data(),
			*offset1	= cdoffset1 ? cdoffset1->real32Data() : 0;

	for (unsigned n=0; n<aduConv->rows(); ++n) {
		unsigned nn = cdgain->elements();
		if (ccdId[n] <= 0 || ccdId[n] > parent->ccds() ||
			nn > parent->nodes())
			errstr<<"Illegal chip and/or node number in row #"
				  <<n<<" of table `"<<aduConv->qualifiedName()<<'\''
			  <<Sas::error("IllegalChipNodeId");

		if (ccdId[n] == ccd) {
			for (unsigned j=0; j<nn; ++j) {
				unsigned p = cdgain->position(n, j);
				//
				//	check for illegal gain values
				//
				if (gain[p] <= 0.)
					errstr<<"gain for CCD #"<<ccdId[n]<<"/node #"<<j
						  <<" is less than or equal to zero"
						  <<Sas::error("IllegalGainValue");

				_offsetConst.push_back(offset0[p]);
				_offsetSlope.push_back(offset1 ? offset1[p] : 0.);
				_gain.push_back(gain[p]);
			}
			break;
		}
	}

	//
	//	check that we have read enough data
	//
	if (_gain.size() != parent->nodes() ||
		_offsetConst.size() != parent->nodes())
		errstr<<"Expected to read "<<parent->nodes()<<" values for the "
				"offset/gain from table `"
			  <<aduConv->qualifiedName()<<"' but actually found: "
			  <<_offsetConst.size()<<"/"<<_gain.size()
			  <<Sas::error("wrongNumberOfElements");

	//	tidy up
	delete cdccdId;
	delete cdgain;
	delete cdoffset0;
	delete cdoffset1;
} // EnergyCorrector::EnergyCorrector()


void EnergyCorrector::dump(ostream &os)
{
	os<<"EnergyCorrector{\n"
	  <<"offset: ";
	for (unsigned n=0; n<_offsetConst.size(); ++n)
		os<<_offsetConst[n]<<'/'<<_offsetSlope[n]<<' ';
	os<<"gain  : ";
	for (unsigned n=0; n<_gain.size(); ++n)
		os<<_gain[n]<<" ";

	os<<"}";
} // EnergyCorrector::dump()


float EnergyCorrector::offsetValue(Node node) const
{
	unsigned n = UNT(node);
	if (n>=_offsetConst.size())
		errstr<<"Unknown readout node "<<INT(node)
			  <<Sas::error("UnknownReadoutNode");

	double deltaT = calServer.state()->exposureStartTime()->get() -
						_offsetRefTime;

	return _offsetConst[n] + deltaT * _offsetSlope[n];
} // EnergyCorrector::offsetValue()


CalReal32Vector &EnergyCorrector::offsetCorrect(const CalReal64Vector &,
											  const CalInt8Vector &nodes,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	errstr<<"Detailed time dependence of offset correction currently not "
			"implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	unsigned nNodes = _offsetSlope.size();
	float offsetVec[nNodes];
	for (unsigned n=0; n<nNodes; ++n)
		offsetVec[n] = offsetValue(Xmm::Node(n));

	float offset;
	for (unsigned n=0; n<eIn.size(); ++n) {
		if (UNT(nodes[n]) < nNodes)
			offset = offsetVec[nodes[n]];
		else {
			errstr<<"No offset value for node "<<INT(nodes[n])
				  <<" event #"<<n<<" - no correction applied"
				  <<Sas::warning("NoSuchValue");
			offset = 0.f;
		}

		eOut[n] = eIn[n] - offset;
	}

	return eOut;
} // EnergyCorrector::offsetCorrect()


CalReal32Vector &EnergyCorrector::offsetCorrect(const CalReal64Vector &time,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	unsigned N=eIn.size();
	CalInt8Vector nodes(0, N);
	int nodeIndex = parent()->node();
	for (unsigned n=0; n<N; nodes[n++] = nodeIndex);

	offsetCorrect(time, nodes, eIn, eOut);

	return eOut;
} // EnergyCorrector::offsetCorrect()


CalReal32Vector &EnergyCorrector::offsetCorrect(const CalReal64Vector &,
											    const CalInt16Vector &,
												const CalInt8Vector &,
												const CalInt16Vector &,
											    CalReal32Vector &eOut) const
{
	errstr<<"Generic energy corrector cannot perform RGS HTR-specific "
			"offset correction"<<Sas::error("UnsupportedOperation");

	return eOut;
} // EnergyCorrector::offsetCorrect()


CalReal32Vector &EnergyCorrector::gainCorrect(const CalReal64Vector &time,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	unsigned N = eIn.size();
	CalInt8Vector nodes(0, N);
	int nodeIndex = parent()->node();
	for (unsigned n=0; n<N; nodes[n++] = nodeIndex);

	gainCorrect(time, nodes, eIn, eOut);
				
	return eOut;
} // EnergyCorrector::gainCorrect()



CalReal32Vector &EnergyCorrector::gainCorrect(const CalReal64Vector &,
											  const CalInt8Vector &nodes,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	errstr<<"Time dependence of gain correction not implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	char nNodes = _gain.size();
	float gain;

	for (unsigned n=0; n<eIn.size(); ++n) {
		if (nodes[n] < nNodes)
			gain = _gain[nodes[n]];
		else {
			errstr<<"No gain value for node "<<INT(nodes[n])
				  <<" event #"<<n<<" - no correction applied"
				  <<Sas::warning("NoSuchValue");
			gain = 1.f;
		}

		eOut[n] = eIn[n];
		if (gain != 0.)
			eOut[n] /= gain;
		else
			errstr<<"Gain is 0 for node "<<INT(nodes[n])<<" on event number "
				  <<n<<" - energy unmodified"
				  <<Sas::warning("ZeroGain");
	}

	return eOut;
} // EnergyCorrector::gainCorrect()


CalReal32Vector &EnergyCorrector::piToeV(const CalReal32Vector &eIn,
										 CalReal32Vector &eOut) const
{
	for (unsigned n=0; n<eIn.size(); ++n)
		// eOut[n] = _eV0 + eIn[n] * _eVperPi;

        /* In practise the conversion from PI to eV is not linear 
           and the current implementation is wrong by a factor of 5.
           Rather than change the header values in many CCF files for
           simplicity we here hardcode the conversion of 1eV per channel.
           This WILL be documented */

        eOut[n] = eIn[n] * 1.0;

	return eOut;
} // EnergyCorrector::piToeV()

CalReal32Vector &EnergyCorrector::eVToPi(const CalReal32Vector &eIn,
                                         CalReal32Vector &eOut) const
{
    for (unsigned n=0; n<eIn.size(); ++n)
        eOut[n] = eIn[n] * 1.0;

    return eOut;
} // EnergyCorrector::eVToPi()

//
//	--------------------------( EMOS )--------------------------------------
//
vector<float> EmosEnergyCorrector::_e1Coef,
			  EmosEnergyCorrector::_e2Coef,
			  EmosEnergyCorrector::_e3Coef,
			  EmosEnergyCorrector::_e4Coef;
bool		  EmosEnergyCorrector::_coefRead = false;


EmosEnergyCorrector::EmosEnergyCorrector(Ccd ccd,
										 const CcfConstituentTable &aduConvTab,
										 const Table *eCombine,
								 CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent),
	  EnergyCorrector(ccd, aduConvTab, parent)
{
	const Table *aduConv = aduConvTab.block();

	//
	//	read offset/gain coefficients
	//
	ColumnData *cdCcdId		= aduConv->column("CCD_ID")->data(),
			   *cdOffset0	= aduConv->column("OFFSET")->data(),
			   *cdOffset1	= aduConv->column("OFFSET1")->data(),
			   *cdOffset2	= aduConv->column("OFFSET2")->data(),
			   *cdOffset3	= aduConv->column("OFFSET3")->data(),
			   *cdGain0		= aduConv->column("GAIN")->data(),
			   *cdGain1		= aduConv->column("GAIN1")->data(),
			   *cdGain2		= aduConv->column("GAIN2")->data(),
			   *cdGain3		= aduConv->column("GAIN3")->data(),
			   *cdPoly0		= aduConv->column("POLY_TERM")->data(),
			   *cdPoly1		= aduConv->column("POLY_TERM1")->data(),
			   *cdPoly2		= aduConv->column("POLY_TERM2")->data(),
			   *cdPoly3		= aduConv->column("POLY_TERM3")->data();

	int8	*ccdId = cdCcdId->int8Data();
	float	*offset0 = cdOffset0->real32Data(),
			*offset1 = cdOffset1->real32Data(),
			*offset2 = cdOffset2->real32Data(),
			*offset3 = cdOffset3->real32Data(),
			*gain0 = cdGain0->real32Data(),
			*gain1 = cdGain1->real32Data(),
			*gain2 = cdGain2->real32Data(),
			*gain3 = cdGain3->real32Data(),
			*poly0 = cdPoly0->real32Data(),
			*poly1 = cdPoly1->real32Data(),
			*poly2 = cdPoly2->real32Data(),
			*poly3 = cdPoly3->real32Data();

	unsigned nodes = parent->nodes();
	if (cdOffset0->elements() != nodes || cdOffset1->elements() != nodes ||
		cdOffset2->elements() != nodes || cdOffset3->elements() != nodes ||
		cdGain0->elements() != nodes || cdGain1->elements() != nodes ||
		cdGain2->elements() != nodes || cdGain3->elements() != nodes ||
		cdPoly0->elements() != nodes || cdPoly1->elements() != nodes ||
		cdPoly2->elements() != nodes || cdPoly3->elements() != nodes)
		errstr<<"Size of vector column in `"<<aduConv->qualifiedName()
			  <<"' does not coincide with number of nodes ("<<nodes<<")"
			  <<Sas::error("CcfDataIntegrityError");

	unsigned n;
	for (n=0; n<aduConv->rows(); ++n)
		if (ccdId[n] == ccd) {
			for (unsigned j=0; j<nodes; ++j) {
				unsigned p = cdOffset0->position(n, j);

				_offsetGain[0][0][j] = offset0[p];
				_offsetGain[0][1][j] = offset1[p];
				_offsetGain[0][2][j] = offset2[p];
				_offsetGain[0][3][j] = offset3[p];

				_offsetGain[1][0][j] = gain0[p];
				_offsetGain[1][1][j] = gain1[p];
				_offsetGain[1][2][j] = gain2[p];
				_offsetGain[1][3][j] = gain3[p];

				_offsetGain[2][0][j] = poly0[p];
				_offsetGain[2][1][j] = poly1[p];
				_offsetGain[2][2][j] = poly2[p];
				_offsetGain[2][3][j] = poly3[p];
			}
			break;
		}

	if (n>aduConv->rows())
		errstr<<"Could not find entry for CCD #"<<ccd<<"in `"
			  <<aduConv->qualifiedName()<<"'"
			  <<Sas::error("CcfDataIntegrityError");

	delete cdCcdId;
	delete cdOffset0;
	delete cdOffset1;
	delete cdOffset2;
	delete cdOffset3;
	delete cdGain0;
	delete cdGain1;
	delete cdGain2;
	delete cdGain3;
	delete cdPoly0;
	delete cdPoly1;
	delete cdPoly2;
	delete cdPoly3;

	//
	//	read FAST_G attribute - see email D. Lumb 2000-11-08
	//
	_fastG = 1.;
	if (aduConv->hasAttribute("FAST_G"))
		_fastG = aduConv->attribute("FAST_G")->asReal();	

	//
	//	read energy combination coefficients
	//
	if (!_coefRead) {
		_coefRead = true;

		ColumnData *cdPattern = eCombine->column("PATTERN_ID")->data(),
				   *cdCoef = eCombine->column("ENERGY_PARAM")->data();
		int8 *pattern = cdPattern->int8Data();
		real32 *coef = cdCoef->real32Data();

		if (cdCoef->elements() != 4)
			errstr<<"Expecting 4 elements in column `"
				  <<cdPattern->parent()->qualifiedName()<<"'"
				  <<Sas::error("CcfDataIntegrityError");

		unsigned N = eCombine->rows();
		_e1Coef.resize(N);
		_e2Coef.resize(N);
		_e3Coef.resize(N);
		_e4Coef.resize(N);
		for (unsigned n=0; n<N; ++n) {
			unsigned patt = pattern[n];
			if (patt >= _e1Coef.size())
				errstr<<"Unexpected pattern ID "<<patt<<" in column `"
					  <<cdPattern->parent()->qualifiedName()<<"'"
					  <<Sas::error("UnknownPattern");
			_e1Coef[patt] = coef[cdCoef->position(n, 0)];
			_e2Coef[patt] = coef[cdCoef->position(n, 1)];
			_e3Coef[patt] = coef[cdCoef->position(n, 2)];
			_e4Coef[patt] = coef[cdCoef->position(n, 3)];
		}

		delete cdPattern;
		delete cdCoef;
	}
} // EmosEnergyCorrector::EmosEnergyCorrector()


//
//	see email D. Lumb, dated 2000-04-13
//
CalReal32Vector &EmosEnergyCorrector::gainCorrect(const CalReal64Vector &,
												  const CalReal32Vector &eIn,
												  const CalInt8Vector &pattern,
												  CalReal32Vector &eOut) const
{
	//
	//	fudge: for CCD1 in modes FastUncompressed/FastCompressed additional
	//	multiplicative factor has to be taken into account
	//
	float effFastG = 1.;
	Mode mode = calServer.state()->mode()->get();
	if (calServer.state()->ccd()->get() == 1 &&
		(mode == Xmm::FAST_UNCOMPRESSED || mode == Xmm::FAST_COMPRESSED))
		effFastG = _fastG;

	int nodeIndx = parent()->node();
	unsigned N = eIn.size();
	for (unsigned n=0; n<N; ++n) {
		unsigned patternIndx = (pattern[n]+3)/4;
		if (patternIndx > 3)
			patternIndx = 3;
		//
		//	pattern ID		patternIndx
		//	0				0
		//	1,2,3,4			1
		//	5,6,7,8			2
		//	9,10,11,12		3
		//	13,...			3	- treat all pattern >12 as pattern [9, 12]
		//

		eOut[n] = _offsetGain[0][patternIndx][nodeIndx] +
				  _offsetGain[1][patternIndx][nodeIndx] * eIn[n] +
				  _offsetGain[2][patternIndx][nodeIndx] * eIn[n] * eIn[n];
		eOut[n] *= effFastG;
	}

	return eOut;	
} // EmosEnergyCorrector::gainCorrect()


//
//	--------------------------( EPN )--------------------------------------
//
bool EpnEnergyCorrector::_gainParamsSet = false;
EpnEnergyCorrector::EpnGainParameters EpnEnergyCorrector::_gainParams;
vector<float> EpnEnergyCorrector::_gainParamsSuppl;
vector<double> EpnEnergyCorrector::_Acoef,
			   EpnEnergyCorrector::_Bcoef;
double EpnEnergyCorrector::_adu_eV,
	   EpnEnergyCorrector::_rawy0;

EpnEnergyCorrector::EpnEnergyCorrector(Ccd ccd,
									   const CcfConstituentTable &aduConvTab,
									   const Table *gainHigh,
									   const Table *reEmission,
									   const Ccf *ccf,
							   CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent),
	  EnergyCorrector(ccd, aduConvTab, parent)
{
	if (!_gainParamsSet) {
		_gainParamsSet = true;

		if (gainHigh->rows() != N_PARAMETERS)
			errstr<<"Expected to find "<<N_PARAMETERS<<" gain parameters "
					"in PN high-accuracy gain table `"
				  <<gainHigh->qualifiedName()<<"' but there are only "
				  <<gainHigh->rows()
				  <<Sas::error("CcfDataIntegrityError");

		const Column *gainCol = gainHigh->column("GAIN_PAR");
		ColumnData *gainParCd = gainCol->data();
		unsigned nElements = gainParCd->elements();
		if (nElements != xmmPayLoad.epn.nNodes() * xmmPayLoad.epn.nCcds())
			errstr<<"Expected to find "<<xmmPayLoad.epn.nNodes() *
										 xmmPayLoad.epn.nCcds()
				  <<" elements in column "<<gainCol->qualifiedName()
				  <<" but there are "<<nElements
				  <<Sas::error("CcfDataIntegrityError");

		real32 *gainPar = gainParCd->real32Data();
	
		//
		//	allocate memory
		//
		for (unsigned n=0; n<N_PARAMETERS; ++n)
			_gainParams.push_back(new float [nElements]);

		for (unsigned n=0; n<N_PARAMETERS; ++n)
			for (unsigned pos=0; pos<nElements; ++pos)
				_gainParams[n][pos] = gainPar[gainParCd->position(n, pos)];

		delete gainParCd;

		//
		//	read supplemental gain parameters
		//
		_gainParamsSuppl.push_back(gainHigh->attribute("FINEPAR1")->asReal());
		_gainParamsSuppl.push_back(gainHigh->attribute("FINEPAR2")->asReal());
		_gainParamsSuppl.push_back(gainHigh->attribute("FINEPAR3")->asReal());

		//
		//	read re-emission threshold parameters
		//
		_adu_eV = reEmission->attribute("ADU_P_EV")->asReal();
		_rawy0 = reEmission->attribute("RAWY0")->asReal();
		CalOalUtility::fillVector(reEmission, "A", 0, _Acoef);
		CalOalUtility::fillVector(reEmission, "B", 0, _Bcoef);
		if (_Acoef.size() != 2 || _Bcoef.size() != 2)
			errstr<<"Expected vector columns A/B to be of size 2 in `"
				  <<reEmission->qualifiedName()<<"'"
				  <<Sas::error("CcfDataIntegrityError");
	}

    /* Get the current observing mode */
	unsigned short modeIndex = calServer.instrument()->
								modeIndex(calServer.state()->mode()->get());
	//
	//	read paramters for additional CTI/gain corrections from CTI CCF
	//
	const XmmInstrument *ins = &XmmPayLoad::epn;
	CifEntry *e = ccf->entry(ins, Ccf::CTI);
	DataSet *set = e->open();

	//	Orsay gain scaling
	Table *tab = set->table("ORSAY_GAIN");
	CalOalUtility::fillVector(tab, "ORSAY_GAIN", _orsayGain);
	if (_orsayGain.size() != ins->nCcds())
		errstr<<"Expected "<<ins->nCcds()<<" Orsay gain parameters in `"
			  <<tab->qualifiedName()<<"'"
			  <<Sas::error("CcfIntegrityError");

	//	long-term CTI increase
     /*  algoid=2 uses a ccd-dependent long-term correction - so have
         to make all long-term vectors per ccd */
	CcfConstituentTable ltCtiTab(set, "LONG_TERM_CTI", 2);
	const Table *lttab = ltCtiTab.block();
	if (ltCtiTab.algoid() == 0) {
		//	ALGOID=0 LT table structure, no-mode dependence; single vector
		//	column if size 5 (0: T-coef, 1-3: E-coef, 4: shift)
		vector<float> temp;
		CalOalUtility::fillVector(lttab, "LONG_TERM_CTI", temp);
		if (temp.size() != 5)
			errstr<<"Expected 5 long-term CTI parameters in `"
				  <<lttab->qualifiedName()<<"'"
				  <<Sas::error("CcfIntegrityError");

        for (unsigned c=1; c<=ins->nCcds(); ++c) {
		  _longTermCtiTcoef[c].assign(4, 0.);	//	3. order polynomial
		  _longTermCtiTcoef[c][1] = temp[0];	//	linear coeff is non-zero
		  _longTermCtiEcoef[c].assign(temp.begin()+1, temp.begin()+4);
		  _longTermCtiShift[c].push_back(temp[4]);
          _longTermCtiCcd.push_back(short(c));
        }

	} else if (ltCtiTab.algoid() == 1){
		//	ALGOID=1 table structure: mode-dependence; three columns
		//	with T/E coefficients + shift value
		vector<short> modes;
		CalOalUtility::fillVector(lttab, "MODE_ID", modes);
		vector<short>::const_iterator i = find(modes.begin(), modes.end(),
											   modeIndex);
		if (i == modes.end())
			errstr<<"Did not find entry matching mode "
				  <<calServer.state()->mode()->get()
				  <<" in "<<lttab->qualifiedName()
				  <<Sas::error("CcfIntegrityError");
		unsigned n=i-modes.begin();

        // Fill a structure for this mode for each CCD
        std::vector<float> tempT, tempE, tempS;
        CalOalUtility::fillVector(lttab, "T_COEFF", n, tempT);
        CalOalUtility::fillVector(lttab, "E_COEFF", n, tempE);
        CalOalUtility::fillVector(lttab, "SHIFT", n, tempS);

        for (unsigned c=1; c<=ins->nCcds(); ++c) {
         _longTermCtiTcoef.push_back(tempT);
         _longTermCtiEcoef.push_back(tempE);
         _longTermCtiShift.push_back(tempS);
         _longTermCtiCcd.push_back(short(c));
        }

	} else if (ltCtiTab.algoid() == 2) {
		//	ALGOID=2 table structure: mode-dependence, ccd-dependence; 
        //           three columns with T/E coefficients + shift value
       vector<short> ccdid, modes;
       CalOalUtility::fillVector(lttab, "CCD_ID", ccdid);
       CalOalUtility::fillVector(lttab, "MODE_ID", modes);

       // Loop over each entry and fill the vectors for the required mode
       short last_ccd=0;
       std::vector<float> tempT, tempE, tempS;

       for (unsigned n=0; n<lttab->rows(); ++n) 
	 {
	   if (modes[n] == modeIndex) {
	     CalOalUtility::fillVector(lttab, "T_COEFF", n, tempT);
	     CalOalUtility::fillVector(lttab, "E_COEFF", n, tempE);
	     CalOalUtility::fillVector(lttab, "SHIFT", n, tempS);
	     _longTermCtiTcoef.push_back(tempT);
	     _longTermCtiEcoef.push_back(tempE);
	     _longTermCtiShift.push_back(tempS);
	     _longTermCtiCcd.push_back(ccdid[n]);
	     
	     // Check that CCDs are in order 
	     if (n>0) assert(ccdid[n]>last_ccd);
	     last_ccd=ccdid[n];
	   }
	 }
	}
	
	// Check that the right number of long term coeffs were in the file
	if (_longTermCtiTcoef[0].size() != 4 ||
	    _longTermCtiEcoef[0].size() != 3 ||
	    _longTermCtiShift[0].size() != 1)
	  errstr<<"Vectors in "<<lttab->qualifiedName()<<" do not have "
	    "expected sizes (4/3/1)"<<Sas::error("CcfIntegrityError");
	
	//	CCD offsets
	tab = set->table("CCD_OFFSETS");
	CalOalUtility::fillVector(tab, "CCD_OFFSETS", _ccdOffsets);
	if (_ccdOffsets.size() != ins->nCcds())
	  errstr<<"Expected "<<ins->nCcds()<<" CCD offset parameters in `"
		<<tab->qualifiedName()<<"'"
		<<Sas::error("CcfIntegrityError");
	
	/*	Temperature gain variations.
	  If available take from QBOXTEMP structure */
	/* Also needed for the old (neves used) correction (AI)*/
	// Extract the gain slope for the required mode
	std::vector<float> gslope;
	    
	if (set->hasBlock("QBOXTEMP_GAIN"))
	  {
	    tab = set->table("QBOXTEMP_GAIN");

	    vector<short> qt_ccdid, qt_modes;
	    CalOalUtility::fillVector(tab, "CCD_ID", qt_ccdid);
	    CalOalUtility::fillVector(tab, "MODE_ID", qt_modes);
	    	    
	    CalOalUtility::fillVector(tab, "SLOPE", gslope);
	    /* Nomimal Quadrant Box  Temperature (AI)*/
	    // Extract the nominal values from the columns
	    std::vector<float> tnom;
	    
	    CalOalUtility::fillVector(tab, "C_REF", tnom);
	    /* Nomimal Quadrant Box  Temperature (AI)*/
	    
	    short last_ccd=0;
	    for (unsigned n=0; n<tab->rows(); ++n) {
	      if (qt_modes[n] == modeIndex) {
		//_TgainSlope.push_back(gslope[n]); 
		_TgainSlope[qt_ccdid[n]]=gslope[n]; 
		/*Nominal Quadrant Box Temperature */
		//_CRef.push_back(tnom[n]);
		_CRef[qt_ccdid[n]]=tnom[n];

		// Check that CCDs are in order 
		if (n>0) assert(qt_ccdid[n]>last_ccd);
		last_ccd=qt_ccdid[n];
	      }
	    }
	    
	  } else { // Use TEMP_CTI block
	  tab = set->table("TEMP_CTI");
	  CalOalUtility::fillVector(tab, "OFFSET", _TgainOffset);	  
	  /* Uses map instead of vector (AI) */
	  //CalOalUtility::fillVector(tab, "SLOPE", _TgainSlope);
	  CalOalUtility::fillVector(tab, "SLOPE", gslope );
	  
	  for (unsigned n=0; n<tab->rows(); ++n) {
	    _TgainSlope[n+1]=gslope[n];
	  }


	}
	// Temperature data common to both structures
	_aduRef = tab->attribute("ADU_REF")->asReal();
	
	// Make sure that all CCDs have temperature slope relation
	// Can't do this because SW, TI,BU modes only have one CCD !
	//if ( _TgainSlope.size() != ins->nCcds())
	//	 errstr<<"Expected "<<ins->nCcds()<<" values in columns of `"
	//		  <<tab->qualifiedName()<<"'"
	//		  <<Sas::error("CcfIntegrityError");

	//	timing mode fine gain
	vector<float> tg, bg, eg;
	tab = set->table("TIMING_GAIN");
	CalOalUtility::fillVector(tab, "TIMING_GAIN", tg);
	tab = set->table("BURST_GAIN");
	CalOalUtility::fillVector(tab, "BURST_GAIN", bg);
	if (tg.size() != 4 || bg.size() != 4)
		errstr<<"Expected "<<ins->nCcds()<<" values in `"
			  <<tab->qualifiedName()<<"' and preceding table"
			  <<Sas::error("CcfIntegrityError");
	_timingGain[0].a1 = tg[0];
	_timingGain[0].a2 = tg[1];
	_timingGain[0].x0 = tg[2];
	_timingGain[0].dx = tg[3];
	_timingGain[1].a1 = bg[0];
	_timingGain[1].a2 = bg[1];
	_timingGain[1].x0 = bg[2];
	_timingGain[1].dx = bg[3];

  /* Gain correction for EFF and FF mode are late additions to the CCF.
     For backwards compatibility only use if the arrays exist in the CCF */
    if (set->hasBlock("EFF_GAIN")) 
    {
	    CcfConstituentTable efgTab(set, "EFF_GAIN", 1);
		tab = set->table("EFF_GAIN");

     /* Read in the EFF_GAIN parameters depending on the ALGOID */
        if (efgTab.algoid() == 0) { // Just one set of params
	  	   CalOalUtility::fillVector(tab, "EFF_GAIN", eg);
        } else { // algoid = 1; Params for each CCD
           vector<unsigned short> ccdid;
           CalOalUtility::fillVector(tab, "CCD_ID", ccdid);

         /* Find the record for this CCD and read the gain values */
           vector<unsigned short>::const_iterator i = find(ccdid.begin(), ccdid.end(),
                                               ccd);
           if (i == ccdid.end())
            errstr<<"Did not find entry matching mode ccd:"
                  <<ccd<<" in "<<tab->qualifiedName()
                  <<Sas::error("CcfIntegrityError");

           unsigned n=i-ccdid.begin();
           CalOalUtility::fillVector(tab, "EFF_GAIN", n, eg);
        }
		_timingGain[2].a1 = eg[0];
		_timingGain[2].a2 = eg[1];
		_timingGain[2].x0 = eg[2];
  		_timingGain[2].dx = eg[3];
    	_effGainRead=true; 
        //cout<< "Gain params: "<<eg[0]<<" "<<eg[1]<<" "<<eg[2]<<" "<<eg[3]
          //  <<endl; 
	} else {
 		_effGainRead=false;
  	}

    if (set->hasBlock("FF_GAIN")) 
    {
		tab = set->table("FF_GAIN");
		CalOalUtility::fillVector(tab, "FF_GAIN", eg);
		_timingGain[3].a1 = eg[0];
		_timingGain[3].a2 = eg[1];
		_timingGain[3].x0 = eg[2];
  		_timingGain[3].dx = eg[3];
    	_ffGainRead=true;
	} else {
 		_ffGainRead=false;
  	}


   /* Rate-dependent CTI correction:
    *
    *   If extension and data exists in the CCF then set the
    *   rate-dep CTI parameters, a0, a1 and a2 from the file. 
    *   Otherwise set to 1.0,0.0,0.0 - which is the NULL correction 
   */
    //vector<float> _rdc_a[3];
    if (set->hasBlock("RATE_DEPENDENT_CTI")) 
    {
       tab = set->table("RATE_DEPENDENT_CTI");

       vector<short> rdc_modes;
       vector<float> vrdc_a0, vrdc_a1, vrdc_a2;

       CalOalUtility::fillVector(tab, "MODE_ID", rdc_modes);
       CalOalUtility::fillVector(tab, "A0", vrdc_a0);
       CalOalUtility::fillVector(tab, "A1", vrdc_a1);
       CalOalUtility::fillVector(tab, "A2", vrdc_a2);

      // Search for the right mode and set coeffs 
       for (unsigned n=0; n<tab->rows(); ++n) {
           if (rdc_modes[n] == modeIndex) {
              if (_rdc_a.size() == 0) {  // Set the first time
                 _rdc_a.push_back(vrdc_a0[n]);
                 _rdc_a.push_back(vrdc_a1[n]);
                 _rdc_a.push_back(vrdc_a2[n]);
              }
            }
       }
    }

   // Push back dummy coeffs which result in no change in case none
   // were set
    if (_rdc_a.size() == 0) {
       _rdc_a.push_back(1.0);
       _rdc_a.push_back(0.0);
       _rdc_a.push_back(0.0);
    }

} // EpnEnergyCorrector::EpnEnergyCorrector()


CalReal32Vector &EpnEnergyCorrector::gainCorrect(const CalReal64Vector &time,
											   const CalInt16Vector &rawX,
											   const CalInt16Vector &rawY,
											   const CalReal32Vector &eIn,
											   CalReal32Vector &eOut) const
{
	if (calServer.state()->accuracyLevel()->get() == LOW) {
		//
		//	low-accurcy gain correction
		//
		EnergyCorrector ec = *this;
		unsigned N = rawX.size();
		CalInt8Vector nodeVec(0, N);
		for (unsigned n=0; n<N; ++n)
			nodeVec[n] = rawX[n]-1;
		ec.gainCorrect(time, nodeVec, eIn, eOut);
		return eOut;
	}

	static bool first = true;

	if (first) {
		first = false;
		//
		//	first time called - need to initialize the PN gain corrector
		//	algorithm
		//
		unsigned N = xmmPayLoad.epn.nNodes() * xmmPayLoad.epn.nCcds();
		float *pp0 = _gainParams[0],
			  *pp1 = _gainParams[1],
			  *pp2 = _gainParams[2],
			  *pp3 = _gainParams[3],
			  *aa = _gainParams[6],
			  *bb = _gainParams[8];
		int M = _gainParamsSuppl.size();
		float suppl[M];
		for (int m=0; m<M; ++m)
			suppl[m] = _gainParamsSuppl[m];

		CALPNALGO_initGain(N, pp0, pp1, pp2, pp3, aa, bb, M, suppl);
	}

	bool mediumAccuracy = calServer.state()->accuracyLevel()->get() == MEDIUM;
	unsigned N = eIn.size();
	Ccd	ccd = parent()->ccd();
	unsigned quadId = EPNcam::toQuadrantId(ccd),
			 ccdId = EPNcam::toCcdId(ccd);
	for (unsigned n=0; n<N; ++n) {
		float	ampr = eIn[n];
		int 	x = INT(rawX[n]) - 1,
				y = INT(rawY[n]) - 1,
				ierr = 0;
		float gainCorrectedAmp, gainCtiCorrectedAmp;

		CALPNALGO_gainCorrect(quadId, ccdId, x, y, ampr, gainCorrectedAmp,
							  gainCtiCorrectedAmp, ierr);

		if (!ierr)
			eOut[n] = mediumAccuracy ? gainCtiCorrectedAmp : gainCorrectedAmp;
		else {
			errstr<<"Gain correction algorithm for event #"<<n
				  <<" (CCD/rawX = "<<ccd<<"["<<quadId<<"/"<<ccdId<<"]/"
				  <<rawX[n]<<") failed - error code: "<<ierr
				  <<" no correction applied - reason for failure: ";
			switch (ierr) {
				case 3: errstr<<"negative amplitude"; break;
				default: errstr<<"unknown"; break;
			}
			errstr<<Sas::warning("gainCorrectionFailed");
			eOut[n] = eIn[n];
		}
	}

	return eOut;
} // EpnEnergyCorrector::gainCorrect()


std::vector<double> EpnEnergyCorrector::additionalGainCtiCorrect(
											CalReal32Vector &ampl,
											const CalInt16Vector &rawY,
											double T,
											bool doTCorr,
											bool doOffsetCorr,
											bool doLongTermCTISoftCorr,
											bool doLongTermCTIyCorr,
											bool doLongTermCTICorr) const
{
	unsigned N = ampl.size();
	vector<double>	corrPar(12);
	const CalState *cs = calServer.state();
	double t = (cs->exposureStartTime()->get() - STime(2000U) +
				.5 * cs->exposureDuration()->get()) /
				STime::secondsPerNonLeapYear;	//	years since 01-01-2000 at
												//	middle of exposure

	Ccd ccd = cs->ccd()->get();
	corrPar[0] = _orsayGain[ccd-1];
	corrPar[1] = t;
	corrPar[2] = _aduRef;
	corrPar[7] = T;

	//
	//	energy rescaling compared to Orsay calibration
	//
	for (unsigned n=0; n<N; ++n)
		ampl[n] *= _orsayGain[ccd-1];

    //    First find the entry for this CCD
    short sccd = ccd;
    
    vector<short>::const_iterator c = 
        find(_longTermCtiCcd.begin(), _longTermCtiCcd.end(), sccd);

    if (c == _longTermCtiCcd.end())
       errstr<<"Did not find entry matching mode ccd:"
             <<ccd<<" in Long-term CTI correction"
             <<Sas::error("CcfIntegrityError");

    unsigned cindex = c - _longTermCtiCcd.begin();

	//
	//	correct for long-term CTE decrease (modeled as gain)
	//
	double			tau = _longTermCtiTcoef[cindex][0] +
						  _longTermCtiTcoef[cindex][1] * t +
						  _longTermCtiTcoef[cindex][2] * t*t +
						  _longTermCtiTcoef[cindex][3] * t*t*t;
	double			lty0 = _longTermCtiShift[cindex][0];

    //cout<<"longterm: "<<ccd<<" "<<cindex<<" "<<_longTermCtiTcoef[cindex][2]<<endl;

	if (doLongTermCTICorr) {
	  corrPar[3] = _longTermCtiTcoef[cindex][1];
	  
	  /* If the energy-dept long-term time-dep correction is required */
	  if (doLongTermCTISoftCorr) {
	    for (unsigned i=0; i<N; ++i) {
	      double logkeV = log10(ampl[i]*5.e-3),
		tt = _longTermCtiEcoef[cindex][0] +
		logkeV*(_longTermCtiEcoef[cindex][1] +
			_longTermCtiEcoef[cindex][2]*logkeV);
	      if (doLongTermCTIyCorr)
		tt *= rawY[i]/lty0;
	      ampl[i] *= 1. + tau*tt;	//	ADUREF = 1
	    }
	    corrPar[8] = _longTermCtiEcoef[cindex][0];
	    corrPar[9] = _longTermCtiEcoef[cindex][1];
	    corrPar[10] = _longTermCtiEcoef[cindex][2];
	  } else { // The non energy-dept long-term time-dep correction
	    for (unsigned i=0; i<N; ++i) {
	      
              /* Calculate the number of Y pixels as the RAW-Y position
                 divided by the long_term_Y_shift. */
	      double tt = doLongTermCTIyCorr ? rawY[i]/lty0 : 1.;
	      
              /* AITOR: this is the change made since the last version.
                 It switches the algorithm back from the experimental
                 one to the original one. */
	      ampl[i] *= 1. + tau*tt;
	      // Apply the long-term correction as a delta to the 
	      // already applied first order correction 
	      //ampl[i] *= pow(( (1.0 - _longTermCtiTcoef[cindex][0]) / 
	      //                       (1. - tau ) ), tt );
	    }
	    corrPar[8] = 0.;
	    corrPar[9] = 0.;
	    
	    corrPar[10] = 0.;
	  }
	  
	  corrPar[11] = doLongTermCTIyCorr ? lty0 : 0.;
	} else {
	  corrPar[3]	= 0.;
	  corrPar[8]	= 0.;
	  corrPar[9]	= 0.;
	  corrPar[10]	= 0.;
	  corrPar[11]	= 0.;
	}

	//
	//	correct for CCD offset (modeled as shift):
	//
	if (doOffsetCorr) {
		for (unsigned i=0; i<N; ++i)
			ampl[i] -= _ccdOffsets[ccd-1];
		corrPar[4] = _ccdOffsets[ccd-1];
	} else
		corrPar[4] = 0.;

	//
	//	correct for temperature variations (modeled as gain):
	//
	if (doTCorr) {
	  map<int,float>::const_iterator it;
	  map<int,float>::const_iterator CRefIT;
	  map<int,float>::const_iterator slopeIT;
	  for (unsigned i=0; i<N; ++i)
	    {
	      /* Change to new linear only (no offset) calculation. 
		 The original calculation was never used, so no need to
		 make backwards compatible. */
	      
	      //ampl[i] *= (1. - T*_TgainSlope[ccd-1]) / _aduRef;

	      /*Now uses maps (AI) */
	      //ampl[i] *= (1. - (T-_CRef[ccd-1])*_TgainSlope[ccd-1]);
	      CRefIT  = _CRef.find(ccd);
	      slopeIT = _TgainSlope.find(ccd);
	      //cout<<"AMPL BEFORE: "<<ampl[i]<<" "<<T<<" "<< CRefIT->second << " "<< slopeIT->second <<endl;
	      /* Final modification. Instead of a temperature they want a constant value (AI) */
	      //ampl[i] *= (1. - (T-tempIT->second)*slopeIT->second);
	      ampl[i] *= (1. - T*slopeIT->second)*CRefIT->second;
	      //cout<<"AMPL AFTER: "<<ampl[i]<< endl;
	    }
	  //ampl[i] *= (1. - (_TgainOffset[ccd-1] + T*_TgainSlope[ccd-1]))/
	  //	_aduRef;
	  corrPar[5] = 0.;
	  //corrPar[5] = _TgainOffset[ccd-1];
	  /* now maps (AI)*/
	  //corrPar[6] = _TgainSlope[ccd-1];
	  corrPar[6] = slopeIT->second;
	  	  	  
	  errstr<<"   Temperature for gain [deg C]  = "<<T<<endl
		<<"   Temperature slope [adu/K]     = "<< slopeIT->second <<endl
		<<"   Temperature reference channel = "<<_aduRef<<endl;
	  
	  
	  /*Just in case the old algorithm is used or not. We
	    have to check that the vector has been filled */
	  if (_CRef.size() != 0)
	    errstr<<"  correction factor to adjust peak position  = " <<  CRefIT->second  <<endl;
	  
	  if (_TgainOffset.size() != 0)
	    errstr <<"   Temperature offset [adu]  = "<<_TgainOffset[ccd-1] <<endl;
	  
	  errstr << Sas::message(Msg::App, Msg::Sparse);
	  /* Print the diagnostic info (AI) */	  	  
	} else {
	  corrPar[5] = 0.;
	  corrPar[6] = 0.;
	}
	
	//	diagnostic output
	if (errHandler.verbosity(Msg::App) >= Msg::Verbose) {
	  errstr<<' '<<Sas::message(Msg::App, Msg::Sparse);
	  errstr<<"   number of CCD to be corrected = "<<ccd<<endl
		<<"   Orsay gain scaling            = "<<_orsayGain[ccd-1]<<endl
		<<"   Years elapsed since 1.1.2000  = "<<t<<endl
		<<Sas::message(Msg::App, Msg::Sparse);
	  if (doLongTermCTICorr) {
	    errstr<<"   CTI increase [1.E-5/yr]       = "
		  <<_longTermCtiTcoef[cindex][1]*1.e5<<endl;
	    if (doLongTermCTISoftCorr) {
	      errstr<<"   energy dependence: coeff(0)  = "
		    <<_longTermCtiEcoef[cindex][0]<<endl
		    <<"   energy dependence: coeff(1)  = "
		    <<_longTermCtiEcoef[cindex][1]<<endl
		    <<"   energy dependence: coeff(2)  = "
		    <<_longTermCtiEcoef[cindex][2]<<endl
		    <<Sas::message(Msg::App, Msg::Sparse);
	    }
	    errstr<<"   reference Y shifts            = "<<lty0<<endl
		  <<Sas::message(Msg::App, Msg::Sparse);
	  }
	  if (doOffsetCorr)
	    errstr<<"   CCD offset [adu]              = "<<_ccdOffsets[ccd-1]
		  <<endl<<Sas::message(Msg::App, Msg::Sparse);
	  
	}

	return corrPar;
} // EpnEnergyCorrector::additionalGainCtiCorrect()


void EpnEnergyCorrector::gainTimingCorrect(CalReal32Vector &ampl) const
{

	unsigned N = ampl.size();

	//
	//	fine-adjustment of the gain in TIMING mode to be applied after all
	//	the gain corrections or immediately before the CTI corrections
	//	the input to the routine is in [adu], the coefficients have been
	//	derived for energies in [eV], therefore scale input by
	//	1 adu = 5.e-0 keV
	//
	double a0 = _timingGain[0].a1 - _timingGain[0].a2;
	for (unsigned n=0; n<N; ++n) {
		double eeV = ampl[n]*5.e-0,
			   h = _timingGain[0].a2 + a0/(1.+exp((eeV-_timingGain[0].x0)/
												_timingGain[0].dx));
		ampl[n] /= h;
	}
} // EpnEnergyCorrector::gainTimingCorrect()


void EpnEnergyCorrector::gainBurstCorrect(CalReal32Vector &ampl) const
{

	unsigned N = ampl.size();

	//
	//	fine-adjustment of the gain in Burst mode to be applied after all
	//	the gain corrections or immediately before the CTI corrections
	//	the input to the routine is in [adu], the coefficients have been
	//	derived for energies in [eV], therefore scale input by
	//	1 adu = 5.e-0 keV
	//
	double a0 = _timingGain[1].a1 - _timingGain[1].a2;
	for (unsigned n=0; n<N; ++n) {
		double eeV = ampl[n]*5.e-0,
			   h = _timingGain[1].a2 + a0/(1.+exp((eeV-_timingGain[1].x0)/
												_timingGain[1].dx));
		ampl[n] *= h;
	}
} // EpnEnergyCorrector::gainBurstCorrect()

void EpnEnergyCorrector::gainEffCorrect(CalReal32Vector &ampl) const
{

    // If Eff gain correction parameters have been read from 
    // the CCF then correct the gain. Otherwise issue a warning.
    if (_effGainRead) {

  		unsigned N = ampl.size();

		//
		//	fine-adjustment of the gain in Eff mode to be applied after all
		//	the gain corrections or immediately before the CTI corrections
		//	the input to the routine is in [adu], the coefficients have been
		//	derived for energies in [eV], therefore scale input by
		//	1 adu = 5.e-0 keV
		//
		double a0 = _timingGain[2].a1 - _timingGain[2].a2;
		for (unsigned n=0; n<N; ++n) {
			double eeV = ampl[n]*5.e-0,
			   h = _timingGain[2].a2 + a0/(1.+exp((eeV-_timingGain[2].x0)/
												_timingGain[2].dx));
			ampl[n] *= h;
		}

    } else { // No gain correction parameters found in CCF
		errstr<<"Gain correction parameters were not present for PN "
                "EFF mode in the CCF. No correction will be applied "
			  <<Sas::warning("NoEFFgainCorrectionFound");
    }

} // EpnEnergyCorrector::gainEffCorrect()

void EpnEnergyCorrector::gainFfCorrect(CalReal32Vector &ampl) const
{

    // If Eff gain correction parameters have been read from 
    // the CCF then correct the gain. Otherwise issue a warning.
    if (_ffGainRead) {

  		unsigned N = ampl.size();

		//
		//	fine-adjustment of the gain in Eff mode to be applied after all
		//	the gain corrections or immediately before the CTI corrections
		//	the input to the routine is in [adu], the coefficients have been
		//	derived for energies in [eV], therefore scale input by
		//	1 adu = 5.e-0 keV
		//
		double a0 = _timingGain[3].a1 - _timingGain[3].a2;
		for (unsigned n=0; n<N; ++n) {
			double eeV = ampl[n]*5.e-0,
			   h = _timingGain[3].a2 + a0/(1.+exp((eeV-_timingGain[3].x0)/
												_timingGain[3].dx));
			ampl[n] *= h;
		}

    } else { // No gain correction parameters found in CCF
		errstr<<"Gain correction parameters were not present for PN "
                "FF mode in the CCF. No correction will be applied "
			  <<Sas::warning("NoFFgainCorrectionFound");
    }

} // EpnEnergyCorrector::gainFfCorrect()

double EpnEnergyCorrector::reEmissionThreshold(double Eadu, int16 rawY) const
{
	double EkeV = Eadu * 1.e-3 * _adu_eV;

	return (_Acoef[0] + _Acoef[1] * EkeV) +
		   (_Bcoef[0] + _Bcoef[1] * EkeV) * (DBL(rawY)-_rawy0);
} // EpnEnergyCorrector::reEmissionThreshold()

std::vector<float> EpnEnergyCorrector::getRateDepGainCoeffs() const
{
 RETURN _rdc_a;
}
      
EpnEnergyCorrector::~EpnEnergyCorrector()
{
	//	free up memory
	for (unsigned n=0; n<_gainParams.size(); ++n) {
		delete _gainParams[n]; _gainParams[n] = 0;
	}
	_gainParams.resize(0);
	_gainParamsSuppl.resize(0);
	_gainParamsSet = false;
} // EpnEnergyCorrector::~EpnEnergyCorrector()


//
//	---------------------( RGS )--------------------------------------------
//
RgsEnergyCorrector::RgsEnergyCorrector(
			Ccd ccd,
			const CcfConstituentTable &aduConvTab,
			const Table *darkFrameTable,
			const Table *offsetOverrideTable,
			CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent),
	  EnergyCorrector(ccd, aduConvTab, parent)
{
	const Table *aduConv = aduConvTab.block();

	//
	//	read rejection/acceptance thresholds and charge pixel
	//	ratios
	//
	_qSplitR[0] = 1.f;	//	by definition
	_qSplitR[1] = aduConv->attribute("QSPLITR2")->asReal();
	_qSplitR[2] = aduConv->attribute("QSPLITR3")->asReal();
	_qSplitR[3] = aduConv->attribute("QSPLITR4")->asReal();

	ColumnData *cdRejThresh	= aduConv->column("REJTHRESH")->data(),
			   *cdAccThresh = aduConv->column("ACCTHRESH")->data();
	real32	*rejThresh = cdRejThresh->real32Data(),
			*accThresh = cdAccThresh->real32Data();

	unsigned n = 0, N = aduConv->rows();
	for (unsigned n=0; n<N; ++n) {
		if (n+1 == ccd) {
			unsigned nn = cdRejThresh->elements();
			for (unsigned j=0; j<nn; ++j) {
				unsigned p = cdRejThresh->position(n, j);
				//
				//	check for negative threshold values
				//
				if (rejThresh[p] < 0. || accThresh[p] < 0.)
					errstr<<"rejection/acceptance threshold for CCD #"
						  <<ccd<<"/node #"<<j
						  <<" is negative"
						  <<Sas::error("CcfDataIntegrityError");
				_rejectionThresholds.push_back(rejThresh[p]);
				_acceptanceThresholds.push_back(accThresh[p]);
			}
			break;
		}
	}
	if (n>=N)
		errstr<<"Did not find rejection/acceptance tresholds for RGS CCD #"
			  <<ccd<<" in `"<<aduConv->qualifiedName()<<'\''
			  <<Sas::error("CcfDataIntegrityError");

	//
	//	try to read RAND_MIN/RAND_MAX from attributes
	//
	_eRandMin = 0.;
	_eRandMax = 1.;	//	default values
	string rmin("RAND_MIN"),
		   rmax("RAND_MAX");
	if (aduConv->hasAttribute(rmin) &&
		aduConv->hasAttribute(rmax)) {
		_eRandMin = aduConv->attribute(rmin)->asReal();
		_eRandMax = aduConv->attribute(rmax)->asReal();
	} else
		if (aduConv->parent()->hasAttribute(rmin) &&
			aduConv->parent()->hasAttribute(rmax)) {
			_eRandMin = aduConv->parent()->attribute(rmin)->asReal();
			_eRandMax = aduConv->parent()->attribute(rmax)->asReal();
		} else
			errstr<<"Did not find "<<rmin<<'/'<<rmax<<" attributes in table "
				  <<aduConv->qualifiedName()<<"' or associated set - "
					"randomization will use default value range ["<<
					_eRandMin<<", "<<_eRandMax<<"]"
				  <<Sas::warning("AttributesNotFound");

	//
	//	read dark frame parameters
	//
	unsigned nNodes = xmmPayLoad.rgs1.nNodes();
	_a0.resize(nNodes);
	_a1.resize(nNodes);
	_a2.resize(nNodes);
	_a3.resize(nNodes);
	_b0.resize(nNodes);
	_b1.resize(nNodes);
	_b2.resize(nNodes);
	_b3.resize(nNodes);
	ColumnData *cda0 = darkFrameTable->column("A_0")->data(),
			   *cda1 = darkFrameTable->column("A_1")->data(),
			   *cda2 = darkFrameTable->column("A_2")->data(),
			   *cda3 = darkFrameTable->column("A_3")->data(),
			   *cdb0 = darkFrameTable->column("B_0")->data(),
			   *cdb1 = darkFrameTable->column("B_1")->data(),
			   *cdb2 = darkFrameTable->column("B_2")->data(),
			   *cdb3 = darkFrameTable->column("B_3")->data(),
			   *cdccd = darkFrameTable->column("CCD_ID")->data(),
			   *cdnode= darkFrameTable->column("NODE_ID")->data();
	real32	*a0 = cda0->real32Data(),
			*a1 = cda1->real32Data(),
			*a2 = cda2->real32Data(),
			*a3 = cda3->real32Data(),
			*b0 = cdb0->real32Data(),
			*b1 = cdb1->real32Data(),
			*b2 = cdb2->real32Data(),
			*b3 = cdb3->real32Data();
	int8	*ccdid = cdccd->int8Data(),
			*nodeid = cdnode->int8Data();

	n = 0;
	N = darkFrameTable->rows();
	bool found = false;
	for (; n<N; ++n)
		if (ccdid[n] == ccd) {
			found = true;
			unsigned node = nodeid[n];
			if (node < nNodes) {
				_a0[node] = a0[n]; _a1[node] = a1[n];
				_a2[node] = a2[n]; _a3[node] = a3[n];
				_b0[node] = b0[n]; _b1[node] = b1[n];
				_b2[node] = b2[n]; _b3[node] = b3[n];
			} else
				errstr<<"Illegal node ID ("<<node<<") seen in `"
					  <<darkFrameTable->qualifiedName()
					  <<Sas::error("CcfDataIntegrityError");
		}

	if (!found)
		errstr<<"Did not find dark frame parameters for RGS CCD #"
			  <<ccd<<" in `"<<darkFrameTable->qualifiedName()<<"'"
			  <<Sas::error("CcfDataIntegrityError");

	delete cda0;
	delete cda1;
	delete cda2;
	delete cda3;
	delete cdb0;
	delete cdb1;
	delete cdb2;
	delete cdb3;
	delete cdccd;
	delete cdnode;

	//
	//	optionally read diagnostic offset data from dynamic CCF constituent
	//	created by rgsoffsetcalc
	//
	_hasDiagnosticOffsetData = offsetOverrideTable;
	if (!_hasDiagnosticOffsetData) {
		flatOffsetCurves();	//	define flat offset curves

		return;
	}

	ColumnData *cdtime = offsetOverrideTable->column("TIME")->data(),
			   *cdmean= offsetOverrideTable->column("MEAN")->data();
    cdccd = offsetOverrideTable->column("CCD_ID")->data(),
	cdnode= offsetOverrideTable->column("NODE_ID")->data();

	real64	*time = cdtime->real64Data();
	ccdid = cdccd->int8Data(),
	nodeid = cdnode->int8Data();
	real32	*mean = cdmean->real32Data();

	vector<double> t[nNodes],
				   offset[nNodes];
	double prevTime[nNodes];
	for(n=0; n<nNodes; prevTime[n++] = -1.);
	n = 0;
	N = offsetOverrideTable->rows();
	for (; n<N; ++n)
		if (UNT(ccdid[n]) == ccd) {
			unsigned node = nodeid[n];
			if (node < nNodes) {
				if (time[n] != prevTime[node]) {
					prevTime[node] = time[n];
					t[node].push_back(time[n]);
					offset[node].push_back(mean[n]);
				} else
					errstr<<"Time in row "<<n<<" of `"
						  <<offsetOverrideTable->qualifiedName()
						  <<"' for CCD "<<ccd<<" node "<<node<<" identical "
							"to previous time - ignoring entry"
						  <<Sas::warning("WrongTime");
			} else
				errstr<<"Illegal node ID ("<<node<<") seen in `"
					  <<darkFrameTable->qualifiedName()
					  <<Sas::error("CcfDataIntegrityError");
		}

	delete cdtime;
	delete cdmean;

	if (!t[0].size() && !t[1].size()) {
		errstr<<"Did not find entry for CCD #"<<ccd<<" in `"
			  <<offsetOverrideTable->qualifiedName()<<"' - offset correction "
				"for this CCD will use CCF-based values"
			  <<Sas::warning("NoDynamicOffsetData");
		unsigned nNodes = xmmPayLoad.rgs1.nNodes();
		for (unsigned node=0; node<nNodes; flatOffsetCurve(node++));

		return;
	}

	for (unsigned node=0; node<nNodes; ++node) {
		if (t[node].size()) {
			//
			//	insert duplicate first/last offset value for three
			//	early/late times times such that the offset curve has a
			//	plateau at the beginning/end
			//
			t[node].insert(t[node].begin(), 100.);
			t[node].insert(t[node].begin(), 10.);
			t[node].insert(t[node].begin(), 0.);
			for (n=0; n<3; ++n)
				offset[node].insert(offset[node].begin(), offset[node][0]);

			double missionDuration = Xmm::missionEndTime -
											Xmm::missionReferenceTime;
			t[node].push_back(missionDuration-1000.);
			t[node].push_back(missionDuration-100.);
			t[node].push_back(missionDuration);
			for (n=0; n<3; ++n)
				offset[node].push_back(offset[node][offset[node].size()-1]);
			_offsetCurves[node] =
				new Slatec::CubicHermiteInterpolant<double>(t[node],
															offset[node]);
		} else {
			//	no data for node <node>
			errstr<<"No dynamic offset data for node "<<node<<" have been "
					"found - using flat offset curve"
				  <<Sas::warning("NoDynamicOffsetData");
			flatOffsetCurve(node);
		}
	}
} // RgsEnergyCorrector::RgsEnergyCorrector()


void RgsEnergyCorrector::flatOffsetCurve(unsigned node)
{
	if (node >= xmmPayLoad.rgs1.nNodes())
		errstr<<"Illegal node ID "<<node<<Sas::error("IllegalNodeId");

	//
	//	define flat offset curve
	//
	vector<double>	timeVec(2),
					offsetVec(2);
	timeVec[0] = 0.;
	timeVec[1] = 1000.;
	offsetVec[0] = offsetVec[1] = offsetValue(Xmm::Node(node));
	_offsetCurves[node] =
			new Slatec::CubicHermiteInterpolant<double>(timeVec, offsetVec);
} // RgsEnergyCorrector::flatOffsetCurve()


void RgsEnergyCorrector::flatOffsetCurves()
{
	unsigned nNodes = xmmPayLoad.rgs1.nNodes();
	for (unsigned n=0; n<nNodes; flatOffsetCurve(n++));
} // RgsEnergyCorrector::flatOffsetCurves()



CalReal32Vector &RgsEnergyCorrector::offsetCorrect(
												 const CalReal64Vector &time,
												 const CalInt16Vector &rawX,
												 const CalInt16Vector &rawY,
												 const CalInt8Vector &shape,
												 const CalInt16Vector &eIn,
												 CalReal32Vector &eOut) const
{
	unsigned N = eIn.size();
	CalInt8Vector nodes(0, N);
	int nodeIndex = parent()->node();
	for (unsigned n=0; n<N; nodes[n++] = nodeIndex);

	offsetCorrect(time, rawX, rawY, shape, nodes, eIn, eOut);

	return eOut;
} // RgsEnergyCorrector::offsetCorrect()


CalReal32Vector &RgsEnergyCorrector::offsetCorrect(
												 const CalReal64Vector &time,
												 const CalInt16Vector &rawX,
												 const CalInt16Vector &rawY,
												 const CalInt8Vector &shape,
												 const CalInt8Vector &nodes,
												 const CalInt16Vector &eIn,
												 CalReal32Vector &eOut) const
{
	randomize(eIn, eOut);
	float offset;
	char nNodes = _offsetConst.size();
	for (unsigned n=0; n<eIn.size(); ++n) {
		if (nodes[n] < nNodes)
			offset = (*_offsetCurves[nodes[n]])(time[n]);
		else {
			errstr<<"No offset value for node "<<INT(nodes[n])
				  <<" event #"<<n<<" - no correction applied"
				  <<Sas::warning("NoSuchValue");
			offset = 0.f;
		}

		eOut[n] -= shapeToGrade(shape[n]) * (offset +
					   darkCurrent(rawX[n], rawY[n], nodes[n]));
	}

	return eOut;
} // RgsEnergyCorrector::offsetCorrect()


//
//	HTR mode offset correction
//
CalReal32Vector &RgsEnergyCorrector::offsetCorrect(
												 const CalReal64Vector &time,
												 const CalInt16Vector &rawX,
												 const CalInt8Vector &nodes,
												 const CalInt16Vector &eIn,
												 CalReal32Vector &eOut) const
{
	CalState *state = calServer.state();

	//	verify we have HTR mode
	Mode mode = state->mode()->get();
	if (mode != Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD &&
		mode != Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD)
		errstr<<"Called RGS HTR-specific offset correction in mode `"
			  <<mode<<"'"<<Sas::error("ImproperMode");

	//	set OCB to 3 + construct dummy rawY/shape vectors
	//	set rawY to middle of readout window in OCB 3; shape is 0
	unsigned oldocb = state->ocb()->get();
	state->ocb()->set(3);

	CalInt16Vector rawY(0, rawX.size());
	int16 v = static_cast<int16>(calServer.instrument()->maxY()/3/2);
	rawY = v;
	CalInt8Vector shape(0, rawX.size());
	shape = 0;

	//	call spectroscopy offset correction
	CalReal32Vector &e =
					offsetCorrect(time, rawX, rawY, shape, nodes, eIn, eOut);
	state->ocb()->set(oldocb);

	return e;
} // RgsEnergyCorrector::offsetCorrect()


float RgsEnergyCorrector::darkCurrent(int16 rawX, int16 rawY, unsigned node)
const
{
	double x = DBL(rawX-1),
		   y = DBL(rawY-1),
		   dc = _a0[node]*exp(_a1[node]*x) +
				_a2[node]*x +
				.5*(_a3[node] + _b3[node]) +
				_b0[node]*exp(_b1[node]*y) +
				_b2[node]*y;	//	see [2], p. 47

	return dc;
} // RgsEnergyCorrector::darkCurrent()


double RgsEnergyCorrector::rejectionThreshold(unsigned grade) const
{
	return rejectionThreshold(grade, calServer.state()->node()->get());
} // RgsEnergyCorrector::rejectionThreshold()


double RgsEnergyCorrector::rejectionThreshold(unsigned grade,
											  Node node) const
{
	if (grade < 1 || grade > NGRADES)
		errstr<<"Passed grade value ("<<grade<<") invalid - must be in "
				"[1, "<<NGRADES<<"]"<<Sas::error("InvalidGradeValue");
	if (node < 0 || node > INT(xmmPayLoad.rgs1.nNodes()))
		errstr<<"Passed node value ("<<node<<") invalid - must be in "
				"[0, "<<xmmPayLoad.rgs1.nNodes()<<"]"
			  <<Sas::error("InvalidNode");

	return _rejectionThresholds[node] / _qSplitR[grade-1];
} // RgsEnergyCorrector::rejectionThreshold()


double RgsEnergyCorrector::acceptanceThreshold(unsigned grade) const
{
	return acceptanceThreshold(grade, calServer.state()->node()->get());
} // RgsEnergyCorrector::acceptanceThreshold()


double RgsEnergyCorrector::acceptanceThreshold(unsigned grade,
											   Node node) const
{
	if (grade < 1 || grade > NGRADES)
		errstr<<"Passed grade value ("<<grade<<") invalid - must be in "
				"[1, "<<NGRADES<<"]"<<Sas::error("InvalidGradeValue");
	if (node < 0 || node > INT(xmmPayLoad.rgs1.nNodes()))
		errstr<<"Passed node value ("<<node<<") invalid - must be in "
				"[0, "<<xmmPayLoad.rgs1.nNodes()<<"]"
			  <<Sas::error("InvalidNode");

	return _acceptanceThresholds[node] / _qSplitR[grade-1];
} // RgsEnergyCorrector::acceptanceThreshold()


vector<float> RgsEnergyCorrector::qSplitRatios() const
{
	vector<float> v;
	for (unsigned n=0; n<NGRADES; v.push_back(_qSplitR[n++]));
	return v;
} // RgsEnergyCorrector::qSplitRatios()



unsigned short RgsEnergyCorrector::shapeToGrade(unsigned short shape) const
{
	//
	//  SHAPE      GRADE
	//    0          1
	//    1          4
	//    2          2
	//    3          2
	//    4          3
	//    5          3
	//    6          3
	//    7          3
	//	 >7          2 + warning
	//
	static unsigned short trans[] = { 1, 4, 2, 2, 3, 3, 3, 3 };

	if (shape <NUM_OF(trans))
		return trans[shape];

	errstr<<"Shape code "<<shape<<" has no associated grade value - shape "
			"codes must be in interval (0, "<<NUM_OF(trans)-1
		  <<"] - returning with grade 2"
		  <<Sas::warning("IllegalNullShape");

	return 2;
} // RgsEnergyCorrector::shapeToGrade()


void RgsEnergyCorrector::randomize(const CalInt16Vector &eIn,
								   CalReal32Vector &eOut) const
{
	calServer.setRandom(_eRandMin, _eRandMax);
	eIn.toReal(eOut);
	calServer.setRandom();
} // RgsEnergyCorrector::randomize()


RgsEnergyCorrector::~RgsEnergyCorrector()
{
	unsigned nNodes = xmmPayLoad.rgs1.nNodes();
	for (unsigned n=0; n<nNodes; ++n)
		delete _offsetCurves[n];
} // RgsEnergyCorrector::~RgsEnergyCorrector()

