//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	GratingDataServer --- the GratingDataServer class
//
//	DESCRIPTION:
//		A class 'GratingDataServer' is defined which provides access to
//		all Grating related calibration quantities:
//			o intercept efficiency
//			o obscuration efficiency
//			o reflection efficiency
//			o self-vignettig efficiency
//			o grating misalignment distribution function
//			o grating bowing distribution function
//			o broadening distribution
//
//	CHANGE HISTORY:
//		UL, 2000-03-10: 1.0.0: - first version
//		UL, 2000-05-30: 1.1.0: - added second form of
//									selfVignettingEfficiency()
//								 that takes directly dispersion angles
//		UL, 2000-06-05: 1.2.0: - removed parameter default values from
//								 methods
//		UL, 2000-07-27: 1.3.0: - added misalignmentFigureDistribution()
//		UL, 2000-08-31: 1.4.0: - added F/L argument to constructor
//		UL, 2000-11-13: 1.5.0: - added reflectionEfficiency(alpha, order)
//		UL, 2000-11-14: 1.6.0: - brought misalignmentFigureDistribution() in
//								 line with [2]
//		UL, 2000-11-17: 1.7.0: - added scatterDistribution(); in line with [2]
//		UL, 2001-01-10: 1.8.0: - added GratingDataServer::chi()
//		UL, 2001-02-16: 1.9.0: - added GratingDataServer::order() to compute
//								 spectral order from alpha/beta/E
//							   - revised I/F of scatteringDistribution()
//							   - new method: scatteringDistributionContribs()
//		UL, 2001-02-20: 1.10.0:- new method: bowingFigureDistribution()
//		UL, 2001-03-12: 1.11.0:- added overloaded version of
//									scatteringDistribution()
//								 which accepts number of output spectrum
//								 channels explicitely
//		UL, 2001-05-04: 1.12.0:- added
//									broadeningDistribution()
//		UL, 2001-05-07: 1.12.1:- updated description of
//									broadeningDistribution()
//		UL, 2002-06-14: 1.13.0:- added inChannelOffset (d/f 0) to
//									misalignmentFigureDistribution()
//									bowingFigureDistribution()
//
//	AUTHOR:
//		U. Lammers (UL), Fri Mar 10 11:39:13 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Data Access Handbook, C. Erd et al.
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/GratingDataServer.h,v 1.33 2008/05/23 10:48:54 aibarra Exp $
//
#ifndef GRATING_DATA_SERVER_H
#define GRATING_DATA_SERVER_H

#include "CalCommon.h"
#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "slatec++.h"
#include "Proxy.h"
#include "Image.h"


//
//	---( GratingDataServer )------------------------------------------------
//
class GratingDataServer : public CalAtom,
	 public virtual CalChild<CalAtomCollection<GratingDataServer> > {
public:
	GratingDataServer(const CcfConstituentTable &misalignmentFigureTable,
					  const CcfConstituentTable &bowingFigureTable,
					  const CcfConstituentTable &scatterTable,
					  const CcfConstituentTable &laScatterTable,
					  const CcfConstituentTable &reflectionTable,
					  const CcfConstituentTable &reflectionCorrTable,
					  const CcfConstituentTable &selfVignCorrTable,
					  const CcfConstituentTable &interceptTable,
					  const CcfConstituentTable &obscurationTable,
					  double alpha0,
					  double grooveSpacing,
					  double FoverL,
					  CalAtomCollection<GratingDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate GratingDataServer object
		//


	virtual void dump(std::ostream &);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	static unsigned id() { return Ccf::GratingData; }
		//
		//	PURPOSE:
		//		Return ID of CalAtom
		//


	virtual InterpolatedCurve interceptEfficiency(double theta,
												  double phi) const;
	virtual InterpolatedCurve obscurationEfficiency(double theta,
												    double phi) const;
		//
		//	PURPOSE:
		//		Return intercept/obscuration efficiency of the RGA at a
		//		particular off-axis/azimuth angle
		//
		//	ARGUMENTS:
		//		theta/phi	: off-axis/azimuth angle [arcsec/rad] at which
		//					  intercept/obscuration efficiency curve is
		//					  requested
		//

	virtual double obscurationEfficiency(double E,
										 double theta,
										 double phi) const;
		//
		//	PURPOSE:
		//		Return obbscuration efficiency for single energy E at
		//		theta/phi
		//

	static const int minOrder = -5;
	static const int maxOrder = 0;
	virtual InterpolatedCurve reflectionEfficiency(double alpha,
												   int order) const;
	virtual InterpolatedCurve reflectionEfficiency(double theta,
												   double phi,
												   int order) const;
		//
		//	PURPOSE:
		//		Return grating reflection efficiency of the RGA as a function
		//		of energy at a particular off-axis+azimuth/incidence angle
		//		and for	a given diffraction order.
		//
		//	ARGUMENTS:
		//		theta/phi	: off-axis/azimuth angle [arcsec/rad] at which
		//					  intercept/obscuration efficiency curve is
		//					  requested
		//		alpha		: grating incidence angle
		//		order		: requested diffraction order in the range
		//						[0, -5]; requesting other orders will
		//					  lead to an error condition
		//


	std::vector<double> selfVignettingEfficiency(const std::vector<double> &
												 dispersionAngles) const;
		//
		//	PURPOSE:
		//		Compute the RGA self-vignetting function at a given
		//		set of dispersion angles
		//
		//	ARGUMENTS:
		//		dispersionAngles:	vector with dispersion angles [rad] at
		//							which to evaluate the self-vignetting
		//							function
		//
		//	RETURNS:
		//		Vector with computed self-vignetting factors corresponding
		//		to input vector <dispersionAngles>
		//


	std::vector<double> selfVignettingEfficiency(const std::vector<double> &E,
												 double theta,
												 double phi,
												 int order) const;
		//
		//	PURPOSE:
		//		Compute the RGA self-vignetting function at a given set
		//		energies via the main grating equation:
		//			order * lambda = d * (cos(beta) - cos(alpha))
		//
		//	ARGUMENTS:
		//		E			: vector with energy values [eV] at which
		//					  self-vignetting efficiency is to be evaluated
		//		theta/phi	: off-axis/azimuth angle [arcsec/rad] at which
		//					  intercept/obscuration efficiency curve is
		//					  requested
		//		order		: requested diffraction order in the range
		//						[0, -5];
		//
		//	RETURNS:
		//		Self-vignetting efficiency data evaluated at given
		//		energies <E>
		//


	Spectrum misalignmentFigureDistribution(
									  double nBetaHWHM,
									  double betaBinWidth,
									  double chromaticMagnification,
									  double inChannelOffset = 0.) const;
	Spectrum bowingFigureDistribution(double nBetaHWHM,
									  double betaBinWidth,
									  double chromaticMagnification,
									  double inChannelOffset = 0.) const;
		//
		//	PURPOSE:
		//		Compute the component of the LSF that is due to the
		//			+ misalignment (misalignmentFigureDistribution)
		//			+ non-flatness/bowing (bowingFigureDistribution)
		//		of the RGA grating plates - see [2] for details
		//
		//	ARGUMENTS:
		//		nBetaHWHM	: number of Half-Width-Half-Maximum units to
		//					  evaluate the RGA distribution to
		//					  this determines the number of channels of the
		//					  output spectrum [rad]
		//		betaBinWidth: width of bins of beta grid [rad]
		//		chromaticMagnification:
		//					  real number that describes the scaling of
		//					  the widths and positions of the (Gaussian)
		//					  distribution's components; normally calculated
		//					  as
		//							sin(alpha)/sin(beta_M)
		//		inChannelOffset:
		//					  offset of the first moment of the underlying
		//					  probability density from the center of the
		//					  returned array; in essence the phase of the
		//					  sampling interval; any offset is permitted,
		//					  but the expected range for this argument is
		//							-.5<=inChannelOffset<+.5
		//
		//	RETURNS:
		//		Normalized probability distribution in beta-space. The first
		//		moment of the distribution coincides with the central
		//		spectrum channel. Please note: The distribution is
		//		not 'wrapped-around' such that its first moment coincides with
		//		channel 0. This is needed to convolve the distribution
		//		with other components to form the LSF. The Spectrum class
		//		has two methods
		//			* Spectrum::firstMoment()
		//			* Spectrum::shift()
		//		to perform this operation efficiently outside
		//		misalignmentFigureDistribution()
		//


	Spectrum scatteringDistribution(double nbetaHWHM,
								    double betaChannelWidth,
									double alpha,
									int order,
									double E,
									bool withSmallAngleContib = true,
									bool withLargeAngleContrib = true) const;
	Spectrum scatteringDistribution(unsigned nChannelsLeft,
									unsigned nChannelsRight,
									double betaChannelWidth,
									double alpha,
									int order,
									double E,
									bool withSmallAngleContib = true,
									bool withLargeAngleContrib = true) const;
		//
		//	PURPOSE:
		//		Compute the component of the LSF that is due to the scattering
		//		of light on the grating plates - see [2] for details;
		//		Please note: The distribution does NOT contain the large,
		//		dominating contribution from the non-scattered light
		//		(delta-function in [2]), hence, it is not normalized to one.
		//		The total integral can be inquired with Spectrum::norm(),
		//		i.e. the non-scattered fraction of the total light is given
		//		by
		//			1 - Spectrum::norm()
		//
		//	ARGUMENTS:
		//		nBetaHWHM	: number of Half-Width-Half-Maximum units to
		//				      evaluate the distribution to; this determines
		//					  the number of channels of the output spectrum
		//		nChannelsLeft/nChannelsRight:
		//					: the number of channels to the left and right
		//					  of the channel containing the beta angle
		//					  <betaM> of the centroid of the unscattered
		//					  light distribution; the total number of
		//					  channels in the output spectrum is thus:
		//						nChannelsLeft+nChannelsRight+1
        //      betaChannelWidth:
        //                    width of channels of beta grid [rad]
		//		alpha		: incidence angle on the gratings [rad]
		//		order		: spectral order index [unitless] in range [-5, 0]
		//		E			: energy of the incident photon [eV]
		//		withSmallAngleContrib, withLargeAngleContrib:
		//					  The scattering distribution is composed of a
		//					  small- and a large angle part (see Eqs. 16-18
		//					  in [2]). Setting either of the parameters to
		//					  'false' will suppress the corresponding
		//					  constituent, i.e., the returned spectrum
		//					  will comprise one component only.
		//
		//	RETURNS:
		//		Probability distribution of the scattered light in beta-space;
		//		The spectrum shall consists of an odd number of channels
		//		with the center of the central channel containing the
		//		beta angle <betaM> of the centroid of the unscattered
		//		light distribution. This is calculated internally from the
		//		input tuple (alpha, order, E) using the main grating dispersion
		//		equation. The size of the output spectrum can be inquired
		//		as spectrum.size(); the channel containing <betaM>
		//		is given as (integer division):
		//			spectrum.size()/2
		//		if <nbetaHWHM> has been specified and
		//			nChannelsLeft
		//		if the first I/F is used.
		//
		//		Please note:
		//			+ The distribution is not 'wrapped-around' such that
		//			  its first moment coincides with channel 0. This is
		//			  needed to	convolve the distribution with other
		//			  components to form the LSF. The Spectrum class has two
		//			  methods
		//				* Spectrum::firstMoment()
		//				* Spectrum::shift()
		//			  to perform this operation efficiently outside
		//			  scatteringDistribution()
		//			+ The below function scatteringDistributionContribs()
		//			  can be used to inquire the relative normalizations
		//			  Fsmall/Flarge (see Eqs. 12/16/17 in [2]) of the
		//			  small and large components after a call to
		//			  scatteringDistribution()
		//


	Spectrum broadeningDistribution(double betaChannelWidth,
									double dLambda,
									double betaM,
									int order) const;
		//
		//	PURPOSE:
		//		Construct spectrum for broadening the LSF to account for
		//		finite width of (incident photon) energy bins.
		//
		//	ARGUMENTS:
        //      betaChannelWidth:
        //                    width of channels of beta grid [rad]
		//		dLambda		: wavelength delta (must be != 0) [Angstrom]
		//		betaM		: nominal beta angle [rad]
		//		order		: spectral order index [unitless] in range [-5, -1]
		//
		//	RETURNS:
		//		Broadening distribution to be used in construction of LSF
		//		through convolution of various components.
		//		Please note: The spectrum is in warp-around,
		//		convolution-ready form and strictly symmetric; always
		//		returns an odd-size array with the axis of reflection at
		//		size()/2.
		//		A null spectrum is returned for dLamda=0 or	order = 0.
		//


	DPair scatteringDistributionContribs() const
		{ return DPair(_Fsmall, _Flarge); }
		//
		//	RETURNS:
		//		A number pair (Fsmall/Flarge) representing the relative
		//		normalization of the small and large angle contributions
		//		to the scattering distribution constructed through a preceding
		//		call to scatteringDistribution().
		//


	double alpha(double theta, double phi) const;
		//
		//	PURPOSE:
		//		Compute grating incidence angle from vector in TELCOORD frame
		//
		//	ARGUMENTS:
		//		theta/phi:	off-axis/azimuth angle in TELCOORD frame ["/rad]
		//


	double chi(double theta, double phi) const;
		//
		//	PURPOSE:
		//		Compute cross dispersion angle from from vector in
		//		TELCOORD frame
		//
		//	ARGUMENTS:
		//		theta/phi:	off-axis/azimuth angle in TELCOORD frame ["/rad]
		//


	double beta(double alpha, int order, double E) const;
	double order(double alpha, double beta, double E) const;
	double energy(double alpha, double beta, int order) const;
	//AI: Wavelength scale. Star
	double energy( double lambda) const;
	//AI: Wavelength scale. End

		//
		//	PURPOSE:
		//		Functions to compute the
		//			dispersion angle/diffraction order/energy
		//		as a function of the respective other quantities by means
		//		of the dispersion grating equation (see [2] for details).
		//
		//	ARGUMENTS:
		//		alpha	: grating incidence angle [rad]
		//		beta	: dispersion angle [rad]
		//		order	: diffraction order [unitless]
		//		E		: photon energy [eV]
		//
		//	RETURNS:
		//		Dispersion angle [rad]/diffraction order [unitless]/
		//		energy [eV].
		//		Please note for order: If the index is not close to an
		//		integer number the input tuple (alpha/beta/E) is unphysical.
		//		


	double alpha0() const { return _alpha0; }
		//
		//	RETURN:
		//		On-axis angle of incidence including RGA rotation about
		//		Y-axis [rad]
		//


	virtual ~GratingDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	//	data
	double				_alpha0,	 //	on-axis angle of incidence [rad]
									 //	including RGA rotation
									 // about Y-axis [rad]
						_grooveSpacing,
									//	distance between adjacent plates [Ao]
						_FoverL,	//	focalLength/|G-M|
						_betaVign;	//	threshold angle for beta vignetting
									//	[rad]

	//
	//	RGA misalignment figure distribution
	//
	std::vector<double>	_ampl[2], _center[2], _sigma[2];
													 //	Gaussian parameters
													 //	of misalignment/bowing
													 // figure distributions

	//
	//	scattering parameters
	//
	std::vector<double>	_scatterA, _scatterB, _scatterRMS, _scatterLcor;
	InterpolatedCurve	_largeAngleScatterRMS;

	mutable double _Fsmall, _Flarge;	//	see Eq. 12/13 in [2]

	InterpolatedCurve	_interceptEfficiency,
						_obscurationEfficiency;
	std::vector<double>	_reflectionEgrid,
						_alphaGrid;
	Image<InterpolatedCurve *>
						_reflectionEfficiencies;	//	for different orders
													//	(X-axis) and energies
													//	(Y-axis)
	std::vector<double>	_reflectionEfficienciesCorrEgrid;
	std::vector<InterpolatedCurve *>
						_reflectionEfficienciesCorrValues;
													//	correction factor to
													//	RGA reflection eff.
													//	as function of E [eV];
													//	one curve for each
													//	order [0, -5]
	InterpolatedCurve	_selfVignettingCorr;		//	correction factor to
													//	self vignetting as
													//	function of beta [rad]
	//	methods

	//	compute RGA self vignetting as function of dispersion angle
	double selfVignetting(double beta) const;

	Spectrum GratingDataServer::gaussianFigureDistribution(
										double nBetaHWHM,
										double betaBinWidth,
										double chromaticMagnification,
										double inChannelOffset,
										const std::vector<double> &amplVal,
										const std::vector<double> &centerVal,
										const std::vector<double> &sigmaVal)
	const;

}; // GratingDataServer

#endif
