//-----------------------------------------------------------------------------
//
//                        XMM Science Analysis System (SAS)
//                       (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//    LowEnergyNoiseServer.cc --- implementation of LowEnergyNoiseServer class
//
//    DESCRIPTION:
//        This implements all methods declared in the LowEnergyNoiseServer class
//
//    CHANGE HISTORY:
//        RDS, 2003-11-21: 1.0.0: - first version
//
//    AUTHOR:
//        R. Saxton
//
//    REFERENCES:
//        [1]: Software Specification Document for the XMM Science Analysis
//             Subsystem, M. Dahlem et. al.,
//
//    $Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/LowEnergyNoiseServer.cc,v 1.2 2007/02/05 15:01:07 rsaxton Exp $
//
#include "CalOalUtility.h"
#include "CalServer.h"
#include "LowEnergyNoiseServer.h"

using namespace CalOalUtility;
using namespace std;

//
//    ---( LowEnergyNoiseServer )------------------------------------------------
//
LowEnergyNoiseServer::LowEnergyNoiseServer(unsigned modeOffset,
                                           unsigned ccdOffset,
                                         const Proxy<SImage> &medianMap,
                                         const Proxy<SImage> &masterOffset,
                                         vector<int8> &rawX,
                                         const vector<SImage> &noiseMap,
                                         const Table *corrValuesTable,
                             CalAtomCollection<LowEnergyNoiseServer> *parent)
    : CalChild<CalAtomCollection<LowEnergyNoiseServer> >(parent),
      _modeOffset(modeOffset), _ccdOffset(ccdOffset), 
      _medianMap(medianMap), _masterOffset(masterOffset), _rawX(rawX), 
      _noiseMap(noiseMap), _hlow(0), _hhigh(0), _corrValue(0) 
{
   /* Get data from the correction value table for this ccd, mode combo */
    ColumnData *cdMode = corrValuesTable->column("MODE")->data(),
               *cdCcd = corrValuesTable->column("CCD")->data(),
               *cdHlow = corrValuesTable->column("H_LOW")->data(),
               *cdHhigh = corrValuesTable->column("H_HIGH")->data(),
               *cdCorValue = corrValuesTable->column("COR_VAL")->data();
    int8    *mode = cdMode->int8Data(),
            *ccdid = cdCcd->int8Data();
    real32  *hlow = cdHlow->real32Data(),
            *hhigh = cdHhigh->real32Data();
    int16   *corrValue = cdCorValue->int16Data();

  // Fill the output arrays with required data
    unsigned N = corrValuesTable->rows();
    
    errstr<<"Found "<<N<<" entries in corrVals table "
          << Sas::message(Msg::App, Msg::Noisy); 

    for (unsigned n=0; n<N; ++n) {                       

        if (modeOffset == mode[n] && ccdOffset == ccdid[n]) {
           errstr<<"Right mode/ccd, entry="<<n<<" hlow="<<hlow[n]
           << Sas::message(Msg::App, Msg::Noisy); 
            _hlow.push_back(hlow[n]);
            _hhigh.push_back(hhigh[n]);
            _corrValue.push_back(corrValue[n]);
        }
    }
    delete cdMode;
    delete cdCcd;
    delete cdHlow;
    delete cdHhigh;
    delete cdCorValue;

} // LowEnergyNoiseServer::LowEnergyNoiseServer()


const std::vector<int8> &LowEnergyNoiseServer::rawxVector() const
{
 //   if (!_rawX)               
      //  errstr<<"No rawx vector for mode/ccd "
         //     <<Xmm::toMode(calServer.instrument()->modes()[_modeOffset])
          //    <<"/"<<_ccdOffset
          //    <<Sas::error("NoRawxVector");    

    return _rawX;
} // LowEnergyNoiseServer::rawxVector()


const std::vector<SImage> &LowEnergyNoiseServer::noiseMapVector() const
{
  //  if (!_noiseMap)
      //  errstr<<"No noise map image for mode/ccd "
             // <<Xmm::toMode(calServer.instrument()->modes()[_modeOffset])
            //  <<"/"<<_ccdOffset
            //  <<Sas::error("NoNoiseMapVector");

    return _noiseMap; 
} // LowEnergyNoiseServer::noiseMapVector()

const SImage &LowEnergyNoiseServer::medianMap() const
{
  //  if (!_medianMap)
    //    errstr<<"No median map image for mode/ccd "
      //        <<Xmm::toMode(calServer.instrument()->modes()[_modeOffset])
        //      <<"/"<<_ccdOffset
         //     <<Sas::error("NoMedianMapVector");

    return *_medianMap;
} // LowEnergyNoiseServer::medianMap()

const SImage &LowEnergyNoiseServer::masterOffset() const
{
  //  if (!_masterOffset)
    //    errstr<<"No master offset image for mode/ccd "
      //        <<Xmm::toMode(calServer.instrument()->modes()[_modeOffset])
        //      <<"/"<<_ccdOffset
         //     <<Sas::error("NoMasterOffsetVector");

    return *_masterOffset;
} // LowEnergyNoiseServer::masterOffset()

const vector<real32> &LowEnergyNoiseServer::lowEnergy() const
{
    return _hlow;
} // LowEnergyNoiseServer::lowEnergy()

const vector<real32> &LowEnergyNoiseServer::highEnergy() const
{
    return _hhigh;
} // LowEnergyNoiseServer::highEnergy()

const vector<int16> &LowEnergyNoiseServer::correctionValue() const
{
    return _corrValue;
} // LowEnergyNoiseServer::correctionValue()


LowEnergyNoiseServer::~LowEnergyNoiseServer()
{
    //delete _hlow;
    // delete _hhigh;
    //delete _corrValue;
} // LowEnergyNoiseServer::~LowEnergyNoiseServer()


CalAtomCollection<LowEnergyNoiseServer>::CalAtomCollection(
                            const XmmInstrument *instrument,
                            const Ccf *ccf,
                            Xmm::Filter filter,
                            AccuracyLevel accuracy)
{
    _filter     = filter;
    _accuracy   = accuracy;
    _nCCDs      = instrument->nCcds();
    _nModes     = instrument->nModes();
    _nEntries   = _nCCDs * _nModes;
    LowEnergyNoiseServer *null = 0;
    _collection = new vector<LowEnergyNoiseServer *>(_nEntries, null);

    CifEntry *e = ccf->entry(instrument, Ccf::Reject);
    DataSet *dataSet = e->open();

   /* The MedianMapIndex extension is used to define the modes and ccds
      represented in the CCF */
    CcfConstituentTable ccftab(dataSet, "MEDIAN_MAP_INDEX");
    const Table * medianMapTab = ccftab.block();
    ColumnData *cdMode = medianMapTab->column("MODE")->data(),
               *cdCcd = medianMapTab->column("CCD")->data(),
               *cdMapName = medianMapTab->column("POINTER")->data();

   /* Get the columns into arrays */
    int8   *mode = cdMode->int8Data(),
           *ccdid = cdCcd->int8Data();

   /* Open the Master Offset map extensions */
    Proxy<SImage> masterOffset;
    const Table * masterOffTab=0;
    int8 *moff_mode=0, *moff_ccdid=0;
    ColumnData *moff_cdMapName=0;

    if (dataSet->hasBlock("MASTER_OFFSET_TABLE_INDEX")) {
       CcfConstituentTable CmastOffTab(dataSet, "MASTER_OFFSET_TABLE_INDEX");
       masterOffTab = CmastOffTab.block();
       ColumnData *moff_cdMode = masterOffTab->column("MODE")->data(),
                  *moff_cdCcd = masterOffTab->column("CCD")->data();
                  moff_cdMapName = masterOffTab->column("POINTER")->data();

      /* Get the integer columns into arrays */
       moff_mode = moff_cdMode->int8Data();
       moff_ccdid = moff_cdCcd->int8Data();
    }

   /* Open the Noise map and correction values extensions */
    CcfConstituentTable CnoiseMapTab(dataSet, "NOISE_MAP_INDEX");
    CcfConstituentTable CcorrValuesTab(dataSet, "CORRECTION_VALUES");
    const Table * noiseMapTab = CnoiseMapTab.block();
    const Table * corrValuesTab = CcorrValuesTab.block();

   /* Read the noise map data */
    ColumnData *noise_cdMode = noiseMapTab->column("MODE")->data(),
               *noise_cdCcd = noiseMapTab->column("CCD")->data(),
               *cdRawx = noiseMapTab->column("RAWX")->data(),
               *cdExposure = noiseMapTab->column("EXPOSURE")->data(),
               *cdNoise = noiseMapTab->column("POINTER")->data();
    int8    *noise_mode = noise_cdMode->int8Data(),  
            *noise_ccdid = noise_cdCcd->int8Data(),
            *rawx = cdRawx->int8Data();
    real32  *exposure = cdExposure->real32Data();
 
   /* Loop over each entry in the MedianMap table */
    //unsigned Nc = medianMapTab->rows();
    for (unsigned i=0; i<medianMapTab->rows(); ++i) {                             
        LowEnergyNoiseServer *&qeds =
                                (*_collection)[indexCcdMode(ccdid[i],mode[i])];

       // Open the median array for this cd, mode combination
        string medianmapName(uppercase(stringColumnRow(cdMapName, i)));
        CcfConstituentArray ccfarr(dataSet, medianmapName, 1);
        ArrayData *ad = ccfarr.block()->data();
        vector<unsigned long> extents = ad->parent()->size();
        Proxy<SImage> medianMap(new SImage(extents[0], extents[1],
                                                         ad->real32Data()));
        delete ad;

        //Proxy<SImage> masterOffset;

       /* Find the Master Offset Map for this ccd,mode combination */
        if (dataSet->hasBlock("MASTER_OFFSET_TABLE_INDEX")) {
          for (unsigned n=0; n<masterOffTab->rows(); ++n) {
            if (moff_ccdid[n] == ccdid[i] && moff_mode[n] == mode[i]) {
               string masteroffName(uppercase(
                                      stringColumnRow(moff_cdMapName, i)));

              // Map the array
               CcfConstituentArray ccfarr(dataSet, masteroffName, 1);
               ArrayData *ad = ccfarr.block()->data();
               vector<unsigned long> extents = ad->parent()->size();
               masterOffset = new SImage(extents[0], extents[1],
                                                         ad->real32Data());
               delete ad;
            }
          }
        } 
        else { /* Set MasterOffset table to zeroes if extension not in CCF.
                  Use the MedianMap array to get the size */
          masterOffset = new SImage(extents[0], extents[1]);
        }

        vector<int8> rawxVector;
        vector<SImage> noisemapVector;
              
       /* Extract noise maps for this ccd,mode combination */
        unsigned Nnc =  noiseMapTab->rows();
        for (unsigned n=0; n<Nnc; ++n) {

            if (noise_ccdid[n] == ccdid[i] && noise_mode[n] == mode[i]) {

              // Add record to the rawx vector
               rawxVector.push_back(rawx[n]);

              /* Get noise map and add to the output vector */
               string noisemapName(uppercase(stringColumnRow(cdNoise, n)));
               CcfConstituentArray ccfarr(dataSet, noisemapName, 1);
               ArrayData *nad = ccfarr.block()->data();
               vector<unsigned long> extents = nad->parent()->size();
        
               errstr<<"Adding noise map for rawx "<<rawx[n]
                     <<" exposure="<<exposure[n]
                     << Sas::message(Msg::App, Msg::Noisy);
               
               SImage nMap(extents[0], extents[1], nad->real32Data());
               
              // Divide the noise map by the exposure time
               for (unsigned x=0; x<extents[0]; x++)
                  for (unsigned y=0; y<extents[1]; y++)
                      nMap[x][y] /= exposure[n];
                      
               noisemapVector.push_back(nMap);
               delete nad;

               errstr<<"Pushed back noise vector "<<n<<" [10,10]="<<nMap[10][10]
                     << Sas::message(Msg::App, Msg::Noisy); 

            } // end of if this mode/ccd block
        } // end of loop over table entries
                      
        qeds = new LowEnergyNoiseServer(mode[i], ccdid[i],
                                           medianMap,
                                           masterOffset,
                                           rawxVector, noisemapVector,
                                           corrValuesTab,
                                           this);
    }

    // Close open column data and set
    delete cdMode; delete cdCcd; delete cdMapName; delete noise_cdMode;
    delete noise_cdCcd; delete cdRawx; delete cdNoise;
    //delete moff_cdMode; delete moff_cdCcd; delete moff_cdMapName;
    
    dataSetServer->close(dataSet);
    
} // CalAtomCollection<LowEnergyNoiseServer>::CalAtomCollection()



LowEnergyNoiseServer *CalAtomCollection<LowEnergyNoiseServer>::select(
                    Ccd ccd,
                    Node,
                    unsigned modeOffset) const
{
    return (*_collection)[indexCcdMode(ccd, modeOffset)];
} // CalAtomCollection<LowEnergyNoiseServer>::select()
