//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmGeoLinearizer --- class related to OM astrometric calibration
//
//	DESCRIPTION:
//		Instances of class OmGeoLinearizer can perform the astrometric
//		correction for the OM detector.
//
//	CHANGE HISTORY:
//		UL, 2000-01-11: 1.2.0: - initial version
//
//	AUTHOR:
//		U. Lammers (UL), Tue Jan 11 16:42:35 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/OmGeoLinearizer.h,v 1.9 2003/11/05 11:01:55 rsaxton Exp $
//

#ifndef OM_GEO_LINEARIZER
#define OM_GEO_LINEARIZER

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalDataVector.h"
#include "Bivariate2DInterpolantScattered.h"

#include <vector>

//
//	---( OmGeoLinearizer )--------------------------------------------------
//
class OmGeoLinearizer : public CalAtom,
			  public virtual CalChild<CalAtomCollection<OmGeoLinearizer> > {
public:
	OmGeoLinearizer(const std::vector<real32> &rawX,
					const std::vector<real32> &rawY,
					const std::vector<real32> &deltaX,
					const std::vector<real32> &deltaY,
					const std::vector<real64> &xa,
					const std::vector<real64> &ya,
					const std::vector<real64> &ixa,
					const std::vector<real64> &iya,
					double plateScale,
					CalAtomCollection<OmGeoLinearizer> *parent = 0);
		//
		//	PURPOSE:
		//		Construct OmGeoLinearizer object.
		//
		//	ARGUMENTS:
		//		rawX/rawY	: coordinates of sample point defining the
		//					  distortion grid; does not need to be
		//					  rectangular
		//		deltaX/deltaY:
		//					  distortion vectors assiciated with the
		//					  grid points defined through rawX/rawY
		//


	static unsigned id() { return Ccf::AstroMet; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(std::ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual void additionalPoints(unsigned nPoints = 4);
		//
		//	PURPOSE:
		//		Set the number of points to be used to calculate
		//		partial derivatives in the below interpolation
		//		routine
		//

	virtual void interpolatedDeltas(const CalReal32Vector &rawX,
									const CalReal32Vector &rawY,
									CalReal32Vector &deltaX,
									CalReal32Vector &deltaY,
									CalReal32Vector &sigmaX,
									CalReal32Vector &sigmaY) const;
		//
		//	PURPOSE:
		//		Compute for a list of input pixel positions in the PIXCOORD1
		//		frame the real offset from the ideal linear pixel grid.
		//		The computation is based on 2-D interpolation of CCF data
		//		sampling the offsets on a pre-defined grid of pixel positions
		//
		//	In:
		//		rawX/rawY	: vector with rawX/rawY coordinates of
		//					  points data points at which the offsets
		//					  are requested
		//	Out:
		//		deltaX/deltaY:interpolated offset vectors corresponding
		//					  to input vectors
		//		sigmaX/sigmaY:uncertainties associated with deltaX/deltaY
		//

	
	virtual void polynomialDeltas(const CalReal32Vector &rawX,
								  const CalReal32Vector &rawY,
								  CalReal32Vector &deltaX,
								  CalReal32Vector &deltaY,
								  CalReal32Vector &sigmaX,
								  CalReal32Vector &sigmaY) const;
		//
		//	PURPOSE:
		//		Compute for a list of input pixel positions in the PIXCOORD1
		//		frame the real offset from the ideal linear pixel grid.
		//		The computation is based on a polynomial fit to the
		//		offset data.
		//
		//	In:
		//		rawX/rawY	: vector with rawX/rawY coordinates of
		//					  points data points at which the offsets
		//					  are requested
		//	Out:
		//		deltaX/deltaY:interpolated offset vectors corresponding
		//					  to input vectors
		//		sigmaX/simgaY:uncertainties associated with deltaX/deltaY
		//

	virtual void inversepolynomialDeltas(const CalReal32Vector &rawX,
								  const CalReal32Vector &rawY,
								  CalReal32Vector &deltaX,
								  CalReal32Vector &deltaY,
								  CalReal32Vector &sigmaX,
								  CalReal32Vector &sigmaY) const;
		//
		//	PURPOSE:
		//		Compute for a list of input pixel positions in the PIXCOORD1
		//		frame the real offset from the ideal linear pixel grid.
		//		The computation is based on a polynomial fit to the
		//		offset data, using an inverse transformation.
		//
		//	In:
		//		rawX/rawY	: vector with rawX/rawY coordinates of
		//					  points data points at which the offsets
		//					  are requested
		//	Out:
		//		deltaX/deltaY:interpolated offset vectors corresponding
		//					  to input vectors
		//		sigmaX/simgaY:uncertainties associated with deltaX/deltaY
		//

	virtual void polycalc(const CalReal32Vector &rawX,
						  const CalReal32Vector &rawY,
								CalReal32Vector &deltaX,
								CalReal32Vector &deltaY,
								CalReal32Vector &sigmaX,
								CalReal32Vector &sigmaY,
	                      const std::vector<real64> &cx,
	                      const std::vector<real64> &cy) const;
		//
		//	PURPOSE:
		//		Compute for a list of input pixel positions in the PIXCOORD1
		//		frame the real offset from the ideal linear pixel grid.
		//		The computation is based on a polynomial fit to the
		//		offset data.
		//
		//	In:
		//		rawX/rawY	: vector with rawX/rawY coordinates of
		//					  points data points at which the offsets
		//					  are requested
        //      cx/cy       : polynomial coefficients to use
        //
		//	Out:
		//		deltaX/deltaY:interpolated offset vectors corresponding
		//					  to input vectors
		//		sigmaX/simgaY:uncertainties associated with deltaX/deltaY
		//


	virtual double plateScale(double) const { return _plateScale; }
		//
		//	PURPOSE:
		//		Return plate scale in [arcsec/pixel]
		//


	virtual ~OmGeoLinearizer();
		//
		//	PURPOSE:
		//		Destruct OmGeoLinearizer object
		//

private:
	//	data members
	Slatec::Bivariate2DInterpolantScattered<float>	*_xoffInter, *_yoffInter;

	std::vector<real64> _ax, _ay, _iax, _iay;
	double			    _plateScale;
};

#endif
