//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmPsfDataServer.h --- An object representing the OM PSF
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2001-02-12: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Tue Feb 13 13:51:06 MET 2001
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Calibration Data Access Handbook, C. Erd et al.
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/OmPsfDataServer.h,v 1.8 2003/10/03 14:15:41 sasbuild Exp $
//

#ifndef OM_PSF_DATA_SERVER_H
#define OM_PSF_DATA_SERVER_H

#include "Image.h"
#include "CalAtomsBase.h"

class OmPsf : public CalAtom,
		public virtual CalChild<CalAtomCollection<OmPsf> > {
public:
	OmPsf(const Table *psfTab,
		  CalAtomCollection<OmPsf> *parent = 0);
		//
		//	PURPOSE:
		//		Construct OmPsf object
		//
		//	ARGUMENTS:
		//		psfTab	: table with radial PSF data
		//


	static unsigned id() { return Ccf::PSF1DRB; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(std::ostream &os) const;
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual double apertureRadius() const { return _apertureRadius; }
		//
		//	RETURNS:
		//		Aperture radius [pixel] used in photometric calibration.
		//


	virtual const OmPsf &forCountFrameRateRatio(double countFrameRateRatio)
		const;
		//
		//	PURPOSE:
		//		Choose count-frame-rate-ratio value for the PSF
		//
		//	ARGUMENTS:
		//		countFrameRateRatio	: count-frame-rate-ratio in interval
		//							  [0, 1]
		//


	virtual const OmPsf &withCenterAt(double cenX, double cenY) const;
		//
		//	PURPOSE:
		//		Define center of PSF in CHIPCOORD frame.
		//
		//	ARGUMENTS:
		//		cenX/cenY	: center of PSF [CHIPCOORD]
		//


	struct OmPsfWindow {
		unsigned	nPixNegX, nPixPosX,
					nPixNegY, nPixPosY;
	};
	virtual SImage image(const OmPsfWindow &psfWindow,
						 unsigned resX = 1, unsigned resY = 1) const;
		//
		//	PURPOSE:
		//		Construct a 2-dimensional image of the OM PSF on a specified
		//		grid. It is supposed that the center of the PSF has been
		//		set through a previous invokation of OmPsf::at(), otherwise
		//		an error condition is raised. The number of pixels of
		//		the PSF image in X/Y direction shall be
		//			(psfWindow.nPixNegX+psfWindow.nPixPosX+1)*xRes
		//								X
		//			(psfWindow.nPixNegY+psfWindow.nPixPosY+1)*yRes
		//		Thus the center of the PSF in the returned image is
		//		at pixel (psfWindow.nPixNegX*xRes, psfWindow.nPixNegY*yRes).
		//		The image is normalized to 1.
		//


	const InterpolatedCurve &encircledEnergy() const;
	const InterpolatedCurve &circleRadius() const;
		//
		//	PURPOSE:
		//		Return an InterpolatedCurve that defines the encircled
		//		energy (EE)-vs-r/r-vs-EE relationship
		//


	~OmPsf();
		//
		//	PURPOSE:
		//		Destruct OmPsf object
		//


private:
	mutable double _cenX, _cenY;		//	center of PSF in CHIPCOORD
	typedef std::pair<std::vector<double>, std::vector<double> >	RadialPsf;
										//	r + intensity
	typedef std::map<double, RadialPsf>	MapType;
	MapType								_radialPsfProfile;
	mutable MapType::const_iterator		_selectedRadialProfile;
	mutable InterpolatedCurve			*_psfr,	//	encircled energy at given
												//	radius
										*_rpsf;	//	radius for given EE

	double								_apertureRadius;
	std::vector<double>					_fStretchCoeff;

	//	methods
	double fStretch() const;
};

#endif

