//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	RgsCoolPixDataServer --- implementation of BadPixel related classes
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 2006-03-31, 1.0.0: - skeleton
//
//	AUTHOR:
//		R. Saxton
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/RgsCoolPixDataServer.cc,v 1.1 2006/04/03 15:56:08 rsaxton Exp $
//

#include <functional>	//	binders
#include <algorithm>
#include <cmath>

#include "CalOalUtility.h"
#include "CalServer.h"
#include "BadPixelDataServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	------------------------( RgsCoolPixDataServer )-----------------------------
//

//	Bad-Pixel-Table code if BPT_CODE attribute not found in CCF;
//	if value changes also change corresponding value in ../f90/caltypes.f90
const string RgsCoolPixDataServer::INVALID_BAD_PIXEL_TABLE_CODE = "0_000_UNK";

RgsCoolPixDataServer::RgsCoolPixDataServer(const string &bptCode,
								CalAtomCollection<RgsCoolPixDataServer> *parent)
	: CalChild<CalAtomCollection<RgsCoolPixDataServer> >(parent),
	  _bptCode(bptCode)
{
} // RgsCoolPixDataServer::RgsCoolPixDataServer()


BadPixelMap RgsCoolPixDataServer::map() const
{
	BadPixelMap m;
	unsigned short modeIndex =
		calServer.instrument()->modeIndex(calServer.state()->mode()->get());
	unsigned N=_badPixels.size();
	for (unsigned n=0; n<N; ++n) {
		vector<short>::const_iterator i =
			find(_badPixels[n].modena.begin(), _badPixels[n].modena.end(),
				 modeIndex);
		if (i != _badPixels[n].modena.end())
			//	pixel is not present in this mode
			continue;
		m[_badPixels[n].loc] = _badPixels[n].bp;
	}

	return m;
} // RgsCoolPixDataServer::map()


RgsCoolPixDataServer::~RgsCoolPixDataServer()
{
} // RgsCoolPixDataServer::~RgsCoolPixDataServer()


CalAtomCollection<RgsCoolPixDataServer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs * _nNodes;
	RgsCoolPixDataServer	*null = 0;
	_collection	= new vector<RgsCoolPixDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::CoolPix);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet, 
								e->queryBlockName(Ccf::toString(Ccf::CoolPix)));
	const Table *table = ccftab.block();
	const string bptCode = optionalAttribute(table, "BPT_CODE",
							 RgsCoolPixDataServer::INVALID_BAD_PIXEL_TABLE_CODE);
	for (unsigned n=0; n<_nEntries;
		(*_collection)[n++] = new RgsCoolPixDataServer(bptCode, this));

	ColumnData *cdccdId	= table->column("CCD_ID")->data(),
			   *cdnodeId = table->column("NODE_ID")->data(),
			   *cdrawx = table->column("RAWX")->data(),
			   *cdrawy = table->column("RAWY")->data(),
			   *cdylen = table->column("YLENGTH")->data(),
			   *cddesc = table->column("DESCRIPTION")->data();
	int8	*ccdId	= cdccdId->int8Data(),
			*nodeId	= cdnodeId->int8Data();
	int16	*rawx	= cdrawx->int16Data(),
			*rawy	= cdrawy->int16Data(),
			*ylen	= cdylen->int16Data();

	bool hasModeCol = table->hasColumn("MODE_N/A"),
		 hasNSatImp = table->hasColumn("N_SAT_IMP");
	ColumnData *cdaux = 0;
	real32	*aux = 0;
	if (hasNSatImp) {
		cdaux = table->column("N_SAT_IMP")->data();
		aux = cdaux->real32Data();
	}

	int minX = instrument->minX(),
		maxX = instrument->maxX(),
		minY = instrument->minY(),
		maxY = instrument->maxY();
	//	loop over all cool pixels
	for (unsigned n=0; n<table->rows(); ++n) {
		if (ccdId[n] <= 0 || ccdId[n] > INT(_nCCDs) ||
			nodeId[n] >= INT(_nNodes))
			errstr<<"Illegal chip and/or node number in row #"
				  <<n<<" of CCF file "
				  <<Xmm::toString(instrument->id())<<"/"
				  <<Ccf::toString(Ccf::BadPix)
				  <<Sas::error("IllegalValues");

		unsigned pos = index(ccdId[n], nodeId[n]);

		//	check that coords are within valid bounds
		if (rawx[n] < minX || rawx[n] > maxX ||
			rawy[n] < minY || rawy[n] < minY)
			errstr<<"Cool pixel location ("<<rawx[n]<<'/'<<rawy[n]
				  <<") read from `"<<table->qualifiedName()
				  <<"' is out of valid bounds ["<<minX<<", "<<maxX
				  <<"]x["<<minY<<", "<<maxY<<"]"
				  <<Sas::error("PixelLocationOutOfBounds");

		PixelLocation			l(rawx[n], rawy[n]);
		BadPixel::DefectTypes	t;
		char flagVal = stringColumnRow(cddesc, n)[0];
		switch(toupper(flagVal)) {
			case 'D'	: t = BadPixel::DEAD; break;
			case 'F'	: t = BadPixel::FLICKERING; break;
			case 'H'	: t = BadPixel::HOT; break;
			case 'P'	: t = BadPixel::PIN_HOLE; break;
			default		: t = BadPixel::ANOMALOUS; break;
		}

		BadPixel bp(l, ylen[n], t, islower(flagVal),
					hasNSatImp ? aux[n] : 0.);
		BadPixelOm ombp(bp, 99);
		const BadPixel &badpix = instrument->id() != Xmm::OM
				? bp
					// this is an OM bad pixel - assign severity as well
					// endcoding has not been decided yet - assign 99 for the
					// moment
				: static_cast<const BadPixel &>(ombp);
		vector<short> modeVec;
		if (hasModeCol) {
			CalOalUtility::fillVector(table, "MODE_N/A", n, modeVec);
			//	erase negative elements which mark the end of the vector
			modeVec.erase(find_if(modeVec.begin(), modeVec.end(),
						  bind2nd(less<short>(), 0)), modeVec.end());
		}

		RgsCoolPixDataServer::LocBadPixModeType c = { l, badpix, modeVec };
		(*_collection)[pos]->_badPixels.push_back(c);
	}

	//	tidy up
	delete cdccdId;
	delete cdnodeId;
	delete cdrawx;
	delete cdrawy;
	delete cdylen;
	delete cddesc;
	if (hasNSatImp)
		delete cdaux;

	dataSetServer->close(dataSet);
} // CalAtomCollection<RgsCoolPixDataServer>::CalAtomCollection()


RgsCoolPixDataServer *CalAtomCollection<RgsCoolPixDataServer>::select(
					Ccd chipNr,
					Node nodeNr,
					unsigned) const
{
	unsigned n = index(chipNr, nodeNr);

	return (*_collection)[n];
} // CalAtomCollection<RgsCoolPixDataServer>::select


#if 0
void BadPixelMapCollection::mergeNodes()
{
	for (unsigned ccd=1; ccd<=_nCCDs; ++ccd) {
		BadPixelMap *m = (*_collection)[index(ccd, 0)];
		if (m)
			for (unsigned node=1; node<_nNodes; ++node) {
				unsigned n = index(ccd, node);
				BadPixelMap *m2 = (*_collection)[n];
				if (m2) {
					//	anachronism: GNU's STL map does not have
					//			     map.insert(iterator, iterator)
					const BadPixelMap::MapType::value_type
											*b = &*m2->_map.begin(),
											*e = &*m2->_map.end();
					m->_map.insert(b, e);
				}
				delete m2;
				(*_collection)[n] = 0;
			}
	}
} // BadPixelMapCollection::mergeNodes
#endif


