//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	ScreenThresholdServer --- implementation of BadPixel related classes
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 2009-04-12: 1.0.0: - first version
//
//	AUTHOR:
//		R. Saxton
//
//	REFERENCES:
//
//

#include <functional>	//	binders
#include <algorithm>
#include <cmath>

#include "CalOalUtility.h"
#include "CalServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	------------------------( ScreenThresholdServer )-----------------------------
//
ScreenThresholdServer::ScreenThresholdServer(const float smin, const float smax,
                            const float hmin, const float hmax,
                            const float rmin, const float rmax,
                            const string & thresh,
								CalAtomCollection<ScreenThresholdServer> *parent)
	: CalChild<CalAtomCollection<ScreenThresholdServer> >(parent),
	  _smin(smin), _smax(smax),_hmin(hmin), _hmax(hmax),
          _rmin(rmin), _rmax(rmax), _thresh(thresh)
{
} // ScreenThresholdServer::ScreenThresholdServer()


ScreenThresholdServer::~ScreenThresholdServer()
{
} // ScreenThresholdServer::~ScreenThresholdServer()

// Return the soft, hard and rate band ranges
const float ScreenThresholdServer::softbandmin() const
                          {return _smin;}

const float ScreenThresholdServer::softbandmax() const
                          {return _smax;}

const float ScreenThresholdServer::hardbandmin() const
                          {return _hmin;}

const float ScreenThresholdServer::hardbandmax() const
                          {return _hmax;}

const float ScreenThresholdServer::ratebandmin() const
                          {return _rmin;}

const float ScreenThresholdServer::ratebandmax() const
                          {return _rmax;}

// Return the threshold
string ScreenThresholdServer::threshold() 
			  {return _thresh;}


CalAtomCollection<ScreenThresholdServer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	ScreenThresholdServer	*null = 0;
	_collection	= new vector<ScreenThresholdServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::ScreenThresh);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet, e->queryBlockName("SCREEN_THRESH"));
	const Table *table = ccftab.block();

	ColumnData *cdccdId	= table->column("CCD_ID")->data(),
	//		   *cdnodeId = table->column("NODE_ID")->data(),
			   *cdsmin = table->column("SOFTBAND_MIN")->data(),
			   *cdsmax = table->column("SOFTBAND_MAX")->data(),
			   *cdhmin = table->column("HARDBAND_MIN")->data(),
			   *cdhmax = table->column("HARDBAND_MAX")->data(),
			   *cdrmin = table->column("RATEBAND_MIN")->data(),
			   *cdrmax = table->column("RATEBAND_MAX")->data(),
			   *cdthrs = table->column("THRESH")->data();

	int8	*ccdId	= cdccdId->int8Data();
			// *nodeId	= cdnodeId->int8Data();

	real32  *softmin = cdsmin->real32Data(),
                *softmax = cdsmax->real32Data(),
                *hardmin = cdhmin->real32Data(),
                *hardmax = cdhmax->real32Data(),
                *ratemin = cdrmin->real32Data(),
                *ratemax = cdrmax->real32Data();

      // Loop over each CCD and set the screenthreshold server
        for (unsigned n=0; n<_nEntries; n++) {
 
            // Find the CCD number and use this as the collection index
                //const string fred = stringColumnRow(cdthrs, n);
                (*_collection)[ccdId[n-1]] = new ScreenThresholdServer(
                        softmin[n], softmax[n], 
                        hardmin[n], hardmax[n],
                        ratemin[n], ratemax[n], 
                        stringColumnRow(cdthrs, n), this);
                       // fred, this);
	}

	//	tidy up
	delete cdccdId;
        delete cdsmin;
        delete cdsmax;
        delete cdhmin;
        delete cdhmax;
        delete cdrmin;
        delete cdrmax;
        //delete cdthrs;

	dataSetServer->close(dataSet);
} // CalAtomCollection<ScreenThresholdServer>::CalAtomCollection()


ScreenThresholdServer *CalAtomCollection<ScreenThresholdServer>::select(
					Ccd ccd,
					Node node,
					unsigned) const
{
	unsigned n = ccd-1; // Index is set as CCD number (-1)

        if (!(*_collection)[n])
             errstr<<"The CCF does not contain a ScreenThreshold data for CCD "
                          <<ccd<<" node "<<node<<Sas::error("NoData");


	return (*_collection)[n];
} // CalAtomCollection<ScreenThresholdServer>::select


