//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	SpecQualityServer.cc --- implementation of miscallaneous CalAtomsCommon classes
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 2009-04-13: 1.0.0: - original
//
//	AUTHOR:
//		R. Saxton, 2009-04-13
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/SpecQualityServer.cc,v 1.1 2009/04/15 15:08:32 rsaxton Exp $
//

#include "SpecQualityServer.h"
#include "CalOalUtility.h"
#include "CalServer.h"

#include <algorithm>
#include <cmath>

using namespace CalOalUtility;
using namespace std;

//
//	----------------------( mode parameters )-----------------------------
//
CalAtomCollection<SpecQualityServer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nModes		= instrument->nModes();
	_nEntries	= _nModes; // Max number of entries possible
	SpecQualityServer *null = 0;
	_collection = new vector<SpecQualityServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::SpecQual);
	DataSet *dataSet = e->open();

      // Depending on the instrument open the correct block
        Table *table = dataSet->table(e->queryBlockName("EMOS1_SPQ"));
        if (instrument->id() == Xmm::EMOS2)
           table = dataSet->table(e->queryBlockName("EMOS2_SPQ"));
        else if (instrument->id() == Xmm::EPN) {
           cout<<"PN"<<endl;
           table = dataSet->table(e->queryBlockName("EPN_SPQ"));
        }
        else if (instrument->id() == Xmm::RGS1)
           table = dataSet->table(e->queryBlockName("RGS1_SPQ_BETA"));
        else if (instrument->id() == Xmm::RGS2)
           table = dataSet->table(e->queryBlockName("RGS2_SPQ_BETA"));

      // Read the arrays from this block
	ColumnData *cdmodeId = table->column("MODE_ID")->data(),
		   *cdngrp = table->column("N_GROUPS")->data(),
		   *cdflag = table->column("SPE_FLAG")->data();

	int8	*modeId = cdmodeId->int8Data();
	int8    *groups = cdngrp->int8Data();

      // The range information
        unsigned N = table->rows();
        vector<vector<short> > ranges;

        for (unsigned n=0; n<N; ++n) {
            vector<short> r0;
            CalOalUtility::fillVector(table, "SPE_FLAG", n, r0);
            ranges.push_back(r0);
        }

      /* Read in the Gain limits arrays */
        Table *gtable = dataSet->table(e->queryBlockName("GAINFIT_LIMIT"));

        ColumnData *cdgmodId = gtable->column("MODE_ID")->data(),
                   *cdinstId = gtable->column("INSTRUMENT_ID")->data(),
                   *cdsmin = gtable->column("GSLOP_MIN")->data(),
                   *cdsmax = gtable->column("GSLOP_MAX")->data(),
                   *cdomin = gtable->column("GOFFS_MIN")->data(),
                   *cdomax = gtable->column("GOFFS_MAX")->data();

	int8	*gmodeId = cdgmodId->int8Data();
        real32  *slope_min = cdsmin->real32Data(),
                *slope_max = cdsmax->real32Data(),
                *off_min = cdomin->real32Data(),
                *off_max = cdomax->real32Data();

      /* Fill in the values for each mode */
	for (unsigned n=0; n<table->rows(); ++n) {

	    if (modeId[n] > _nModes) {
                cout<<"NMODES: "<<_nModes<<"NO: "<<modeId[n]<<endl;;
		errstr<<"Illegal mode number in row #"
			  <<n<<" of CCF file "
			  <<Xmm::toString(instrument->id())<<"/"
			  <<Ccf::toString(Ccf::ModeParam)
			  <<Sas::error("IllegalValues");
            }

	    //unsigned pos = index(modeId[n]);

         // Find the gain limits entry corresponding to this mode
            short gl = -999;
            for (unsigned g=0; g<gtable->rows(); ++g) 
                if ( Xmm::toInstrument(stringColumnRow(cdinstId, g)) == 
                        instrument->id() && gmodeId[g] == modeId[n])
                            gl=g;
            
           /* Some instruments and/or modes may not have gain limit values 
              so if not found set to the first value without a warning */
            if (gl == -999) gl = 0;

         // Set the values for this mode
	    (*_collection)[n] = new SpecQualityServer;

	    (*_collection)[n]->_ngroups = groups[n];
	    (*_collection)[n]->_ranges = ranges[n];
	    (*_collection)[n]->_gslope_min = slope_min[gl];
	    (*_collection)[n]->_gslope_max = slope_max[gl];
	    (*_collection)[n]->_offset_min= off_min[gl];
	    (*_collection)[n]->_offset_max= off_max[gl];
	}

      // Delete the column pointers
        delete cdmodeId;
        delete cdngrp;
        delete cdflag;
	delete cdgmodId;
	delete cdsmin;
	delete cdsmax;
	delete cdomin;
	delete cdomax;

	dataSetServer->close(dataSet);
} // CalAtomCollection<SpecQualityServer>::CalAtomCollection


SpecQualityServer *CalAtomCollection<SpecQualityServer>::select(
					Ccd,
					Node,
					unsigned modeOffset) const
{
	unsigned n = index(modeOffset);

	return (*_collection)[n];
} // CalAtomCollection<SpecQualityServer>::select
